package algebraTree;
import java.util.*;
import DSRG.MW.*;

public class Fold extends SchemaSQLOperator {
    /** a vector of strings denoting attributes that are to be folded */
    Vector attributeNames = new Vector(5);
    String relationName;
    String pivotName, dataName;
    
    /** relSchema is the list of attributes in the output relation
        that are neither pivotName nor dataName nor in attributeNames */
    Vector relSchema;

    /** constructs a non-leaf Fold-operator. No local source is
        needed. The vector "attributes" contains all the attributes
        which should be folded.  */
    public Fold(DBManager localDB, Vector attributes,
                String relation, String pivot, String data) {
        super(localDB,NOLEAF); 
        setSourceDB(localDB); // the source is an intermediate relation
        attributeNames = attributes;
        relationName = relation;
        pivotName = pivot;
        dataName = data;
    }
    
    /** constructs a leaf Fold-operator. You must supply a sourceDB */
    public Fold(DBManager localDB, DBManager sourceDB, 
                Vector attributes, String relation, 
                String pivot, String data) {
        super(localDB,LEAF); 
        if (!setSourceDB(sourceDB)) {
            System.out.println("Fold operator wants to be a leaf.");
            System.exit(0);
        }
        attributeNames = attributes;
        relationName = relation;
        pivotName = pivot;
        dataName = data;
    }
    
    public String getPivotName() {
        return pivotName;
    }

    public String getDataName() {
        return dataName;
    }

    protected Vector computeLocalQueries() {
        Vector result =  new Vector(2);
        String query = new String();
        boolean firstTime = true;
        relSchema = sourceDB.getAttributeNames(relationName);
        // need to remove all element in relSchema which are also in
        // attributeNames, since that will be the remaining schema for
        // the output relation.
        int index;
        for (Enumeration enum=attributeNames.elements();
             enum.hasMoreElements();) {
            index=relSchema.indexOf(enum.nextElement());
            if (index>=0)
                relSchema.removeElementAt(index);
        }
        
        // now we can flatten the vector out to a string since we'll
        // be making queries:
        String selectClause = new String();
        for (Enumeration enum=relSchema.elements();
             enum.hasMoreElements();)
            selectClause += (", "+ (String)enum.nextElement());
        
        // and now we generate queries.
        String pivotAtt;
        for (Enumeration enum=attributeNames.elements();
             enum.hasMoreElements();) {
            pivotAtt = (String)enum.nextElement();
            if (firstTime)
                firstTime = false;
            else {
                query = query + " UNION ";
            }
            query += "select "
                + "'"+ pivotAtt +"'"
                + " as "
                + "\""+pivotName+"\""
                + ", " + pivotAtt
                + " as "+ dataName
                + selectClause
                + " from "+relationName;
        }
        result.addElement(query);
        if (Main.DEBUG) System.out.println("FOLD produced: "
                                           +result.toString());
        return result;
    }


    /** propagate an update through this operator */
    /* package private */Vector propagateUpdate(Update upd) {
        if (Main.DEBUG) 
            System.err.println(this.toString()+": Trying to propagate: "+upd);
 
       // if update is not for me, return null
        Vector result = new Vector();
        String updPivot;
        if (!upd.relation.equalsIgnoreCase(relationName)) return null;

        if (upd instanceof DataUpdate) {
            DataUpdate updD = (DataUpdate)upd;
            // a data update
            if (updD.before==null) { // add-tuple
                // relSchema contains all attributes that are not Ap/Ad and not int attributeNames.
                // so relSchema plus pivotName plus dataName are the schema of the output relation.
                Vector outputSchema=(Vector)relSchema.clone();
                // this reflects the order in the select clause in computeLocalQueries()
                outputSchema.insertElementAt(dataName,0);
                outputSchema.insertElementAt(pivotName,0);
                //now pivotName is at position 0, dataName at position 1!!
                Table insertTable = new Table();
                insertTable.setRelName("DUMMY");
                insertTable.setAttributeNames(outputSchema);
                Vector insertValues = new Vector(outputSchema.size());
                // two dummy values as placeholders for the pivotNAme and dataName
                insertValues.addElement("dummy value");
                insertValues.addElement("dummy value");
                //then, copy relevant values from old update into new update template
                Vector dataValuesFromUpdate = ((Tuple)updD.after.getTuples().elementAt(0)).getVector();
                for (int i = 2; i < outputSchema.size(); i++)
                    insertValues.addElement(dataValuesFromUpdate.elementAt(updD.after.getAttributeNames().indexOf(((String)outputSchema.elementAt(i)).toUpperCase())));
                // create a tuple
                Tuple insertTuple = new Tuple(insertTable, insertValues);
                // and create a table with it. This will be our template.
                insertTable.addTuple(insertTuple);
                DataUpdate tmpUpdate;
                Table tmpTable;
                String currentAttribute;
                for (Enumeration enum=attributeNames.elements();
                     enum.hasMoreElements();) {
                    currentAttribute = (String)enum.nextElement();
                    // create a new table. The copy constructor
                    // properly copies (duplicates) the values so they
                    // don't point to the same objects!
                    tmpTable = new Table(insertTable); 
                    // now set pivotElement
                    ((Tuple)tmpTable.getTuples().elementAt(0)).getVector().setElementAt(currentAttribute,0);
                    // and the data element
                    updPivot =
                        (String)updD.after.getColumn(currentAttribute).elementAt(0);
                    ((Tuple)tmpTable.getTuples().elementAt(0)).getVector().setElementAt(updPivot,1);
                    // and use this tmpTable to create another data update
                    tmpUpdate = new DataUpdate(outputTableName,null,tmpTable);
                    result.addElement(tmpUpdate);
                }
            }
            else if (updD.after==null) {// delete tuple
                
                Table deleteTable = executeUpdateQuery(outputTableName, relSchema, updD.before);
                if (deleteTable.getTuples().size()>0)
                    result = QENode.mergeVectors(result,tableToDeleteUpdates(outputTableName,deleteTable));
                else // invalid delete-tuple
                    Util.exit("Invalid update. Tuple to be deleted doesn't exist.");
            }
            else { // update-tuple

                // this recursive solution is simpler but less
                // efficient. May change sometime (yeah,right!!)
                // 082700 koeller

                // recursive call. Make sure we don't have update-tuple-changes here!!!!
                // add-tuple
                result = this.propagateUpdate(new DataUpdate(updD.relation,null,updD.after));
                // delete-tuple
                result = QENode.mergeVectors(result,this.propagateUpdate(new DataUpdate(updD.relation,updD.before, null)));
            }
        }
        else {
            SchemaUpdate updS = (SchemaUpdate)upd;
            SchemaUpdate tmpSchemaUpdate;
            // a schema change
            if (updS.attribute!=null) {
                // an attribute update
                switch (updS.type) {
                case SchemaUpdate.ADD:
                    if (attributeNames.indexOf(updS.attribute.toUpperCase())!=-1) {
                        // this can never happen!!! all this work for
                        // nothing :( 

                        //A \in u_1....u_n query INPUT relation and
                        //for each tuple construct an output tuple

                        // we flatten out the relSchema Vector
                        /*
                        String selectClause = new String();
                        for (Enumeration enum=relSchema.elements();
                             enum.hasMoreElements();)
                            selectClause += (", "+ (String)enum.nextElement());
                        
                        String query = "select "
                            + "'"
                            + updS.attribute 
                            +"'"
                            + " as "
                            + "\""+pivotName+"\""
                            + ", " + "null"
                            + " as "+ dataName
                            + selectClause
                            + " from "+relationName;
                        
                        Table inputTable = sourceDB.executeViewQuery(query);
                        // (without where-clause)

                        DataUpdate tmpUpdate;
                        Table tmpTable;
                        // now iterate through table and construct new DataUpdate for each tuple
                        for (Enumeration enum=inputTable.getTuples().elements();
                             enum.hasMoreElements();) {
                            tmpTable = new Table();
                            tmpTable.setRelName("DUMMY");
                            tmpTable.addTuple((Tuple)enum.nextElement());
                            tmpUpdate = new DataUpdate(outputTableName,null,tmpTable);
                            result.addElement(tmpUpdate);
                        }
                        */
                    }
                    else { //A something else
                        tmpSchemaUpdate = new SchemaUpdate(updS);
                        tmpSchemaUpdate.relation = new String(outputTableName);
                        result.addElement(tmpSchemaUpdate);
                    }
                    relSchema.addElement(updS.attribute);
                    
                    break;

                case SchemaUpdate.DELETE:
                    if (attributeNames.indexOf(updS.attribute.toUpperCase())==-1) {
                        // attribute not in Fold-List
                        tmpSchemaUpdate = new SchemaUpdate(updS);
                        tmpSchemaUpdate.relation = new String(outputTableName);
                        result.addElement(tmpSchemaUpdate);
                        relSchema.removeElement(updS.attribute);
                    }
                    else {
                        updPivot =
                            (String)updS.attribute;
                        Table deleteTable = localDB.executeViewQuery("select * from "+outputTableName+" where "+pivotName+" = '"+updPivot.toUpperCase()+"'"); // don't look at capitalization
                        result = QENode.mergeVectors(result,tableToDeleteUpdates(outputTableName,deleteTable));
                        attributeNames.removeElement(updPivot);
                    }
                    break;
                case SchemaUpdate.RENAME:
                    // the SchemaUpdate class converts the newName
                    // attribute into capitals. That is necessary for
                    // this to work.
                    if (attributeNames.indexOf(updS.attribute.toUpperCase())==-1) {
                        // propagate
                        tmpSchemaUpdate = new SchemaUpdate(updS);
                        tmpSchemaUpdate.relation = new String(outputTableName);
                        result.addElement(tmpSchemaUpdate);
                        relSchema.setElementAt(updS.newName,relSchema.indexOf(updS.attribute));
                    }
                    else { // update Q set A_p=b where A_p=a;
                        updPivot =
                            (String)updS.attribute;
                        Table updateTable = localDB.executeViewQuery("select * from "+outputTableName +" where "+pivotName+" = '"+updPivot.toUpperCase()+"'");
                        DataUpdate tmpUpdate;
                        Table tmpTable,tmpTable2;
                        Tuple tmpTuple;
                        // now iterate through table and construct new DataUpdate for each tuple
                        int pivotIndex = updateTable.getAttributeNames().indexOf(pivotName);
                        for (Enumeration enum=updateTable.getTuples().elements();
                             enum.hasMoreElements();) {
                            tmpTable = new Table();
                            tmpTable.setRelName("DUMMY");
                            tmpTable.setAttributeNames(updateTable.getAttributeNames());
                            tmpTuple = (Tuple)enum.nextElement();
                            tmpTable.addTuple(tmpTuple);
                            tmpTable2 = new Table(tmpTable);
                            ((Tuple)tmpTable2.getTuples().elementAt(0))
                                .getVector().setElementAt(updS.newName,pivotIndex);
                            tmpUpdate = new DataUpdate(outputTableName,tmpTable,tmpTable2);
                            result.addElement(tmpUpdate);
                        }
                        attributeNames.setElementAt(updS.newName,attributeNames.indexOf(updS.attribute));
                    }
                }
            }
            else {
                switch (updS.type) {
                case SchemaUpdate.ADD:
                    // do nothing... makes no sense
                    break;
                case SchemaUpdate.DELETE:
                    tmpSchemaUpdate = new SchemaUpdate(updS);
                    tmpSchemaUpdate.relation = new String(outputTableName);
                    result.addElement(tmpSchemaUpdate);
                    break;
                case SchemaUpdate.RENAME:
                    relationName = updS.newName;
                    break;
                }
            }
        }
        if (Main.DEBUG) System.out.println("Incremental Update Propagation. FOLD gave us: "+result.toString());
        // the QENode above this must execute these updates!!!
        return result;
    }




}
