package algebraTree;

import java.util.Vector;
import java.util.Enumeration;
import DSRG.MW.*;

public class QENode {

    /** the associated Operator object */
    private SchemaSQLOperator operator;

    /** Vector of QENodes. This is a n-ary tree */
    private Vector children = new Vector();

    /** constructs a QENode with an existing operator attached to
        it. Children have to be added explicitly using the addChild()
        method. Therefore, the tree has to be built starting from the
        leaves! */

    public QENode(SchemaSQLOperator myOperator) {
        operator = myOperator;
    }
    


    // this tells all the children to compute their results and then
    // computes its own
    public void computeResult() {

        // depth-first tree traversal
        Enumeration enum = children.elements();
        while (enum.hasMoreElements())
            ((QENode)enum.nextElement()).computeResult();
        operator.executeQuery();
        /*
        // now drop all local tables... changed 082100, now done in finalize()
          enum = children.elements();
          while (enum.hasMoreElements())
          ((QENode)enum.nextElement()).operator.dropLocalTable();
        */
    }
    
    /** kicks off update propagation of a certain update. The update
        concerns a certain relation which has to be found (and its
        operator has to be identified). If this function is called in
        any other than the root node of the algebra tree, the query
        evaluation will not be correct. */
    public void executeUpdate(Update upd) {
        Vector updates = recursiveUpdate(this, upd);
        // for testing purposes
        if (Main.DEBUG) System.out.println("update propagation yields: \n"+updates);
        //        getOperator().executeUpdates(updates);
    }

    /** propagates upd all the way down to the leaves and then
     *  recursively collects updates up the operator tree. Each leaf
     *  must be able to decide if the update affects it or not.
     *  @return result which contains a vector of vectors. Each
     *  sub-vector is a list of updates from a certain node. The
     *  affected tables/attributes are stored in each update. */
    private Vector recursiveUpdate(QENode tmpNode, Update upd) {
        Vector result=new Vector();
        Vector intermediateResult=new Vector();
        Vector updatesFromOneNode;
        // base case
        SchemaSQLOperator tmpOperator = tmpNode.getOperator();
        if (tmpOperator.isLeaf) {
            updatesFromOneNode = tmpOperator.propagateUpdate(upd);
            result = mergeVectors(result,updatesFromOneNode);
        }
        else {
            for (Enumeration enum=tmpNode.children.elements();
                 enum.hasMoreElements();) {
                updatesFromOneNode = recursiveUpdate(((QENode)enum.nextElement()),upd);
                intermediateResult = mergeVectors(intermediateResult,updatesFromOneNode);
            }
            
            //System.err.println("QENode::recursiveUpdate.intermediateResult:\n"+intermediateResult);
            
            // and now channel the updates through to the parent node
            for (Enumeration enum=intermediateResult.elements();enum.hasMoreElements();) {
                updatesFromOneNode = tmpOperator.propagateUpdate((Update)enum.nextElement());
                result = mergeVectors(result,updatesFromOneNode);
            }
        }
        if (Main.DEBUG) System.out.println(tmpOperator.getClass().getName()+" is executing updates in "+tmpOperator.outputTableName);
        tmpOperator.executeUpdates(result);
        return result;
    }

    /** a helper function that merges two Vectors */
    /* package private */ static Vector mergeVectors(Vector a, Vector b){ 
        Vector result = a;
        Object curElement;
        if (b!=null && b.size()>0)
            for (Enumeration enum=b.elements();enum.hasMoreElements();) {
                curElement = enum.nextElement();
                if (curElement!=null)
                    result.addElement(curElement);
            }
        return result;
    }

    /** asks this node to drop its local table (mainly used for the
        root node)*/
    public boolean dropLocalTable() {
        return operator.dropLocalTable();
    }

    /** returns the name of the local relation of this node */
    public String getRelName() {
        return operator.getRelName();
    }

    /** sets the name if this node's operator. Make sure that you use
        this only for the root node (not essential, but it's not
        needed for any other purpose */
    public void setRelName(String s) {
        operator.setRelName(s);
    }

    /** needed by Builder::findFlatAttName2 */
    public SchemaSQLOperator getOperator() {
        //        if (!Main.DEBUG) System.out.println("QENode::getOperator() should not be used in production version!");
        return operator;
    }
    
    /** return a child. Should not be used except for that one hack in
        the treeBuilder */
    public QENode getChild(int index) {
        if (index>=children.size()) return null;
        return (QENode)children.elementAt(index);
    }
    
    /**  add a child node to the current Node */
    public boolean addChild(QENode baby) {
        if (operator.isLeaf) {
            System.err.println("Leaf nodes can not have children");
            return false;
        }
        else {
            children.addElement(baby);
            return true;
        }
    }

    /** this function prints the algebra tree in a somewhat primitive fashion */
    public void printTree(String indent) {
        System.out.println(indent+this.toString());
        if (children.size()==0)
            return;
        else {
            Enumeration enum = children.elements();
            QENode current;
            while (enum.hasMoreElements()) {
                current  = (QENode)enum.nextElement();
                current.printTree(indent+"  ");
            }
        }
    }

    
    public String toString() {
        String result;
        result= "QENode containing a "+operator.getClass().getName()+"-operator";
        if (operator instanceof StandardSQL) 
            result += "  Query is : "+((StandardSQL)operator).queryString;
        return result;
    }
}



