package algebraTree;
import java.util.*;
import DSRG.MW.*;

public class Split extends SchemaSQLOperator {
    /** the name of the input table */
    String relationName;
    
    /** the name of the pivot attribute to split on */
    String pivotName;

    /** the relational schema of the input and output tables, without
     * the dynamic content. This is used for both query evaluation and
     * incremental propagation. */
    Vector relSchema = null;

    /** the value in the pivot attribute in an update. */
    String updPivot;

    /** constructs a non-leaf Split-operator. No local source is needed */
    public Split(DBManager localDB,
                String relation, String pivot) {
        super(localDB,NOLEAF); 
        setSourceDB(localDB); // the source is an intermediate relation
        relationName = relation;
        pivotName = pivot;
    }
    
    /** constructs a leaf Split-operator. You must supply a sourceDB */
    public Split(DBManager localDB, DBManager sourceDB, String relation, 
                String pivot) {
        super(localDB,LEAF); 
        if (!setSourceDB(sourceDB)) {
            System.out.println("Split operator wants to be a leaf.");
            System.exit(0);
        }
        relationName = relation;
        pivotName = pivot;
    }
    
    protected Vector computeLocalQueries() {
        Vector result =  new Vector(10);
        String query = new String();
        boolean firstTime = true;
        String selectClause = new String();
        relSchema = sourceDB.getAttributeNames(relationName);
        // need to remove pivotName and dataName from relSchema since
        // that will be the remaining schema for the output relation.
        relSchema.removeElementAt(relSchema.indexOf(pivotName));
        
        // now we can flatten the vector out to a string since we'll
        // be making queries:
        for (Enumeration enum=relSchema.elements();
             enum.hasMoreElements();) {
            if (firstTime)
                firstTime = false;
            else
                selectClause += ", ";
            selectClause += (String)enum.nextElement();
        }        

        // now we retrieve all distinct values of the pivot attribute
        // and their counts
        pivotData = 
            sourceDB.executeViewQuery("select "
                                      + pivotName
                                      + ", count("
                                      + pivotName
                                      + ") from "
                                      + relationName
                                      + " group by "
                                      + pivotName);

        Vector pivotElements = pivotData.getTuples();

        String pivotAtt;
        
        // now generate one select statement for each table
        for (Enumeration enum=pivotElements.elements();
             enum.hasMoreElements();) {
            pivotAtt = (String)((Tuple)enum.nextElement()).getVector().elementAt(0);
            query = ("select "
                      + selectClause
                      + " from "
                      + relationName
                      + " where "
                      + pivotName
                      + " = '"+pivotAtt+"'"
                      );
            
        result.addElement(query);
        }
        if (Main.DEBUG) System.out.println("Split produced: "+result.toString());
        return result;
    }

    protected boolean executeQuery() {
        Vector queries = computeLocalQueries();
        Table result;

        String currentQuery;
        String currentPivot;
        int startPos;
        boolean stillGood = true;
        
        // we need to take each query, parse the pivot string out
        // again, and create a table with that name.
        for (Enumeration enum=queries.elements();
             enum.hasMoreElements();) {
            currentQuery = (String)enum.nextElement();
            startPos = currentQuery.indexOf('\'')+1;
            currentPivot = 
                currentQuery.substring(
                                       startPos, 
                                       currentQuery.indexOf('\'',startPos));
            if (Main.DEBUG) System.out.println("Current Query: "+currentQuery);
            result = sourceDB.executeViewQuery(currentQuery);
            result.setRelName(currentPivot);
            stillGood = localDB.storeRelation(result);
            if (!stillGood) break;
        }
        
        return stillGood;
    }


    /** propagate an update through this operator */
    /* package private */Vector propagateUpdate(Update upd) {
        //        if (Main.DEBUG) 
        //  System.err.println(this.toString()+": Trying to propagate: "+upd);
        // if update is not for me, return null

        Vector result = new Vector();
        if (!upd.relation.equalsIgnoreCase(relationName)) return null;

        if (upd instanceof DataUpdate) {
            DataUpdate updD = (DataUpdate)upd;
            // a data update
            if (updD.before==null) { // add-tuple
                updPivot =
                    ((String)updD.after.getColumn(pivotName).elementAt(0)).trim();
                int pivotCount = findPivotCount(updPivot);
                if (pivotCount>0) {// A* doesn't change
                    incPivotCount(updPivot);
                }
                else { // A* changes
                    addPivot(updPivot);
                    // this is an add-rel!! there is no schema, that is a bug (or feature)
                    result.addElement(new SchemaUpdate(updPivot,SchemaUpdate.ADD));
                }
                // then translate update query into vector of data
                // updates

                // The tuple should not be there, so we don't check it.

                DataUpdate tmpUpdate;

                // create add-tuple-update with the new schema
                Table tmpTable = updD.after;
                Vector schema = tmpTable.getAttributeNames();
                Vector values = ((Tuple)tmpTable.getTuple(0)).getVector();
                for (int i = 0; i < schema.size(); i++) {
                    if (((String)schema.elementAt(i))
                        .equalsIgnoreCase(updPivot)) {
                        values.removeElementAt(i);
                        schema.removeElementAt(i);
                        break;
                    }
                }
                tmpUpdate = new DataUpdate(updPivot, null,new Table(tmpTable));
                result.addElement(tmpUpdate);
            }
            else if (updD.after==null) {// delete tuple
                updPivot = 
                    ((String)updD.before.getColumn(pivotName).elementAt(0)).trim();
                int pivotCount = findPivotCount(updPivot);
                if (pivotCount==1) { // delete that relation
                    delPivot(updPivot);
                    SchemaUpdate tmpUpdate = new SchemaUpdate(updPivot,SchemaUpdate.DELETE);
                    result.addElement(tmpUpdate);
                }
                else if (pivotCount>1) { // update. remove corresponding tuple
                    decPivotCount(updPivot);
                    Table tmpTable = executeUpdateQuery(updPivot, relSchema, updD.before);
                    if (tmpTable==null)
                    Util.exit("Invalid update. Tuple to be deleted doesn't exist.");
                    DataUpdate tmpUpdate = new
                        DataUpdate(updPivot,new Table(tmpTable),null);
                    result.addElement(tmpUpdate);
                }
                else // invalid delete-tuple
                    Util.exit("Invalid update. Tuple to be deleted doesn't exist.");
            }
            else { // update-tuple
                // recursive call. Make sure we don't have update-tuple-changes here!!!!
                // add-tuple
                DataUpdate tmpUpdD= new DataUpdate(updD);
                result = this.propagateUpdate(new DataUpdate(tmpUpdD.relation,null,tmpUpdD.after));
                // delete-tuple ( after add since otherwise relation
                // would be temporarily empty and would be deleted)
                result = QENode.mergeVectors(result,this.propagateUpdate(new DataUpdate(tmpUpdD.relation,tmpUpdD.before,null)));
            }
        }
        else {
            SchemaUpdate updS = (SchemaUpdate)upd;
            SchemaUpdate tmpUpd;
            // a schema change
            if (updS.attribute!=null) {
                // an attribute update
                switch (updS.type) {
                case SchemaUpdate.ADD:
                    //                    for (Enumeration enum=((Tuple)pivotData.getTuple(0)).getVector().elements();
                    for (Enumeration enum=pivotData.getColumn(0).elements();
                         enum.hasMoreElements();) {
                        tmpUpd = new SchemaUpdate(updS);
                        tmpUpd.relation = (String)enum.nextElement();
                        result.addElement(new SchemaUpdate(tmpUpd));
                    }
                    // maintain relation schema vector
                    relSchema.addElement(updS.attribute);
                    break;
                case SchemaUpdate.DELETE:
                    // deleting pivot attribute doesn't work
                    if (updS.attribute.equalsIgnoreCase(pivotName)) {
                        Util.invalidView();
                    }
                    else
                        // okay, so we need to propagate this att-del
                        // to all output relations.
                        //                        for (Enumeration enum=((Tuple)pivotData.getTuples().elementAt(0)).getVector().elements();
                    for (Enumeration enum=pivotData.getColumn(0).elements();
                             enum.hasMoreElements();) {
                            tmpUpd = new SchemaUpdate(updS);
                            tmpUpd.relation = (String)enum.nextElement();
                            result.addElement(new SchemaUpdate(tmpUpd));

                        }
                    // and maintain relSchema (BUG 101000)
                    relSchema.removeElementAt(relSchema.indexOf(updS.attribute));


                    break;
                case SchemaUpdate.RENAME:
                    if (updS.attribute.equalsIgnoreCase(pivotName)) {
                        pivotName = updS.newName;
                    } 
                    else {
                        //                        for (Enumeration enum=((Tuple)pivotData.getTuples().elementAt(0)).getVector().elements();
                    for (Enumeration enum=pivotData.getColumn(0).elements();
                             enum.hasMoreElements();) {
                            tmpUpd = new SchemaUpdate(updS);
                            tmpUpd.relation = (String)enum.nextElement();
                            result.addElement(new SchemaUpdate(tmpUpd));
                        }
                        // maintain relSchema
                        relSchema.setElementAt(updS.newName,relSchema.indexOf(updS.attribute));
                    }
                }
            }
            else {
                switch (updS.type) {
                case SchemaUpdate.ADD:
                    // do nothing... makes no sense
                    break;
                case SchemaUpdate.DELETE:
                    for (Enumeration enum=((Tuple)pivotData.getTuples().elementAt(0)).getVector().elements();
                         enum.hasMoreElements();) {
                        tmpUpd = new SchemaUpdate(updS);
                        tmpUpd.relation = (String)enum.nextElement();
                        result.addElement(new SchemaUpdate(tmpUpd));
                    }
                    break;
                case SchemaUpdate.RENAME:
                    relationName = updS.newName;
                    break;
                }
            }
        }
        if (Main.DEBUG) System.out.println("Incremental Update Propagation. SPLIT gave us: "+result.toString());
        // the QENode above this must execute these updates!!!
        return result;
    }

    /** This function takes a vector of updates and applies it to the
     outputRelation of this operator.  This is overridden for the
     Split-Class. Must be called AFTER propagate Update */
    /* package private*/ void executeUpdates(Vector updates) {

        //        if (updPivot==null) Util.exit("updPivot is null. This is a bug.");
        
        Update upd;
        DataUpdate updD; SchemaUpdate updS;
        String updTable;
        boolean firstTime;
        String query = new String();
        for (Enumeration enum = updates.elements();
             enum.hasMoreElements();) {

            upd = (Update)enum.nextElement();
            if (upd instanceof DataUpdate) {
                updD = (DataUpdate)upd;
                updTable = updD.relation;
                //if (Main.DEBUG) System.out.println("Split node - updating table: "+updTable);
                
                if (updD.before == null) {//add-tuple
                    //insert-query
                    String names = new String();
                    String values = new String();
                    String tmpValue;
                    for (int i = 0; i<updD.after.getAttributeNames().size(); i++) {
                        tmpValue = (String)updD.after.getAttributeNames().elementAt(i);
                        if (!tmpValue.equalsIgnoreCase(pivotName)) {
                            if (i>=1) { names += ","; values += ","; }
                            names += tmpValue;
                            tmpValue = (String)((Tuple)updD.after.getTuples().elementAt(0)).getVector().elementAt(i);
                            if (tmpValue==null) values += "null";
                            else values += "'"+tmpValue+"'";
                            
                        }
                    }
                    query = "insert into "
                        + updTable
                        + " ("
                        + names
                        + ") values("
                        + values 
                        + ")";
                    int result = localDB.executeSQLUpdate(query);
                }
                else { // delete- or update-tuple
                    // construct where-clause
                    String whereClause = new String();
                    String tmpValue;
                    for (int i = 0; i<updD.before.getAttributeNames().size(); 
                         i++) {
                        
                        tmpValue = (String)updD.before.getAttributeNames().elementAt(i);
                        if (!tmpValue.equalsIgnoreCase(pivotName)) {
                            if (i>=1) whereClause += " and ";
                            whereClause += tmpValue;
                            tmpValue = (String)((Tuple)updD.before.getTuples().elementAt(0)).getVector().elementAt(i);
                            if (tmpValue==null) whereClause += " is null";
                            else whereClause += "='"+tmpValue+"'";
                        }
                    }
                    // now whereClause contains a where clause for all tuple values in the before-tuple
                    if (updD.after==null) {//delete-tuple
                        query = "delete from "
                            + updTable
                            + " where "
                            + whereClause;
                        int result = localDB.executeSQLUpdate(query);
                        if (result==0)
                            Util.exit("Tuple "+whereClause+" to be deleted not found... weird...");
                        
                    }
                    else { // update-tuple
                        // set-query
                        String setClause = new String();
                        String tmpEntry,tmpValueBefore;
                        boolean notThisOne;
                        for (int i = 0; i<updD.before.getAttributeNames().size(); i++) {
                            tmpEntry=new String();
                            tmpValue = (String)updD.after.getAttributeNames().elementAt(i);
                            if (tmpValue.equalsIgnoreCase(pivotName)) {
                                if (i>=1) tmpEntry += ", ";
                                tmpEntry += tmpValue + "=";
                                tmpValue = (String)((Tuple)updD.after.getTuples().elementAt(0)).getVector().elementAt(i);
                                tmpValueBefore = (String)((Tuple)updD.before.getTuples().elementAt(0)).getVector().elementAt(i);
                                notThisOne=false;
                                if (tmpValue==null) {
                                    if (tmpValueBefore!=null) tmpValue = "null";
                                    else notThisOne=true;
                                }
                                else {
                                    if (tmpValue.equals(tmpValueBefore)) tmpValue = "'"+tmpValue+"'";
                                    else notThisOne=true;
                                }
                                if (!notThisOne) setClause += tmpEntry + tmpValue;
                            }
                        }
                    // now whereClause contains a where clause for all
                    // tuple values in the before-tuple and setClause
                    // contains all values to be set which are
                    // different from before

                        query = "update "
                            + updTable
                            + " set "
                            + setClause
                            + " where "
                            + whereClause;
                        int result = localDB.executeSQLUpdate(query);
                        if (result==0)
                            Util.exit("Tuple to be updated not found... weird..."+query);
                    } // else from if (updD.after==null)
                } // else from if(updD.before==null)
            }  // if upd instanceof DataUpdate
            else {
                updS = (SchemaUpdate)upd;
                switch (updS.type) {
                case SchemaUpdate.ADD:
                    if (updS.attribute==null) { // relation-update

                        // we must assume here that the schema is the
                        // same as the one of any other output
                        // relation.  we'll use Oracle-specific SQL
                        // and copy from pivotNameNames(0)... bad hack,
                        // but whatever (092500)

                        localDB.deleteRelation(updS.relation,true);
                        localDB.executeSQLUpdate(
                                                 "create table "
                                                 + updS.relation
                                                 + " as select * from "
                                                 + ((Tuple)pivotData.getTuple(0)).getVector().elementAt(0) + " where rownum<=1");
                        localDB.executeSQLUpdate("delete from "+updS.relation);
                    }
                    else {//attribute-update
                        System.out.println("WARNING: adding attribute with dummy data type char(31)");
                        localDB.addAttribute(updS.relation,updS.attribute,"CHAR",31);
                    }
                    break;
                case SchemaUpdate.DELETE:
                    if (updS.attribute==null) { // relation-update
                        localDB.deleteRelation(updS.relation,false); // non-silent delete
                    }
                    else {//attribute-update
                        localDB.deleteAttribute(updS.relation,updS.attribute);
                    }
                    break;
                case SchemaUpdate.RENAME:
                    if (updS.attribute==null) { // relation-update
                        localDB.renameRelation(updS.relation,updS.newName);
                    }
                    else {//attribute-update
                        localDB.renameAttribute(updS.relation,updS.attribute,updS.newName);
                    }
                    break;
                default:
                    Util.exit("invalid schema update");
                }
            }
        }
    }

}
