package algebraTree;
import java.util.*;
import DSRG.MW.*;

public class StandardSQL extends SchemaSQLOperator {
    String queryString;

    private static final int ADD=1;
    private static final int DELETE=2;

    /** constructs a non-leaf StandardSQL-operator. 
        No local source is needed */
    public StandardSQL(DBManager localDB, String query) {
        super(localDB,NOLEAF); 
        setSourceDB(localDB); // the source is an intermediate relation
        queryString = query;
    }
    
    /** constructs a leaf StandardSQL-operator. You must supply a sourceDB */
    public StandardSQL(DBManager localDB, DBManager sourceDB, String query) {
        super(localDB,LEAF); 
        if (!setSourceDB(sourceDB)) {
            System.out.println("StandardSQL operator wants to be a leaf.");
            System.exit(0);
        }
        queryString = query;
    }
    
    protected Vector computeLocalQueries() {
        Vector result =  new Vector(1);
        
        result.addElement(queryString);
        return result;
    }

     
    /** propagate an update through this operator.  To decide which
        updates to propagate, we actually store each tuple in a
        relation, execute the view query against this relation, and
        retrieve the result.*/
    Vector propagateUpdate(Update upd) {
        //        System.err.println(this.toString()+": StandardSQL is trying to propagate: "+upd);
        Vector result = new Vector();
        if (upd instanceof DataUpdate) {
            DataUpdate updD = (DataUpdate)upd;
            // a data update. We need to decide about the update
            // propagation here. It's not trivial.
            //    updD.relation = outputTableName;
            //    result.addElement(new DataUpdate(updD));

            /* So: we create a table with the schema of the update
             (both tables should have the same schema). Then, we run
             this operator's query against that table and see what
             happens.  */

            //add-tuple: as above. If count is >=1, we propagate the tuples
            //delete-tuple: again, as above.
            if (updD.before==null) {// add-tuple
                QENode.mergeVectors(result,propagateTuples(updD,ADD));
            }
            else if (updD.after==null) {//delete-tuple
                QENode.mergeVectors(result,propagateTuples(updD,DELETE));
            }
            else {//update-tuple
                // update tuple is broken down into an add first, then
                // a delete. That's all we can do for now.
                QENode.mergeVectors(result,propagateTuples(updD,ADD));
                QENode.mergeVectors(result,propagateTuples(updD,DELETE));
            }
        }
        else {
            SchemaUpdate updS = (SchemaUpdate)upd;
            // a schema change
            updS.relation = outputTableName;
            result.addElement(new SchemaUpdate(updS));
        }

        // now if we don't actually have any updates propagated, we return null (good idea?):
        if (result.size()==0) 
            return null;
        else {
            if (Main.DEBUG) System.out.println("Incremental Update Propagation. StandardSQL gave us: "+result.toString());
            return result;
        }
    }

    /** this function takes an update, stores it in a table and then
        applies this operator's query to that table. The result is
        returned as a Vector of DataUpdates which can then be
        propagated. Slow but works. 092800: changed the logic, by
        adding a nested SQL statement to the original query that
        replaces the original relation.*/

    private Vector /*of DataUpdates*/ propagateTuples(DataUpdate update, int propagationMode) {
        Vector result = null;
        Table propagatedTuple;
        if (propagationMode==ADD)
            propagatedTuple = update.after;
        else
            propagatedTuple = update.before;
        

        /* Now comes a trick: we will replace the relation that's
           being updated with a nested select-clause that contains the
           update tuple (and thus creates a relation containing only
           that one tuple */
        String nestedSelect = "(select ";
        Vector attNames = propagatedTuple.getAttributeNames();
        Vector attValues = propagatedTuple.getTuple(0).getVector();
        String tmpValue;
        for (int i = 0 ; i < attNames.size(); i++) {
            if (attValues.elementAt(i)==null)
                tmpValue = "null";
            else 
                tmpValue = "'"+attValues.elementAt(i)+"'";
            nestedSelect +=
                tmpValue
                + " as "
                + attNames.elementAt(i);
            if (i<attNames.size()-1) 
                nestedSelect += ",";
            else
                nestedSelect += " from dual)";
        }

        String tmpQuery = new String(queryString);
        String updatedTableName = update.relation.toUpperCase();
        int index = tmpQuery.toUpperCase().indexOf("FROM");
        
        while (index != -1) {
            // find occurence of updatedTableName in from-clause
            index = tmpQuery.toUpperCase().indexOf(updatedTableName,index);
            if (index==-1) break;
            tmpQuery = tmpQuery.substring(0,index)
                + " "
                + nestedSelect
                + " "
                + tmpQuery.substring(index);
            // move index to after the current from-clause
            index += nestedSelect.length()+3;
        }

        Table tmpTable = localDB.executeViewQuery(tmpQuery);

        Vector tuples = tmpTable.getTuples(); 
        if (tuples!=null) {
            result = new Vector();
            Tuple currentTuple; Table currentTable;
            for (Enumeration enum=tuples.elements(); enum.hasMoreElements();) {
                currentTuple = (Tuple)enum.nextElement();
                currentTable = new Table(1);
                currentTable.setRelName(outputTableName);
                currentTable.setAttributeNames(tmpTable.getAttributeNames());
                // PROBLEM: the current tuple's parent table will now be wrong
                currentTable.addTuple(currentTuple);
                if (propagationMode==ADD)
                    result.addElement(new DataUpdate(outputTableName, null, currentTable));
                else
                    result.addElement(new DataUpdate(outputTableName, currentTable, null));
            }
        }
        return result;
    };

}

