package algebraTree;
import java.util.*;
import DSRG.MW.*;

public class Unfold extends SchemaSQLOperator {
    /** this is the INPUT-TABLE NAME*/
    String relationName;
    String pivotName, dataName;

    /** the relational schema of the input and output tables, without
     * the dynamic content. This is used for both query evaluation and
     * incremental propagation. */
    Vector relSchema = null;

    /** constructs a non-leaf Unfold-operator. No source DB is needed */
    public Unfold(DBManager localDB,
                String relation, String pivot, String data) {
        super(localDB,NOLEAF); 
        setSourceDB(localDB); // the source is an intermediate relation
        relationName = relation;
        pivotName = pivot;
        dataName = data;
    }
    
    /** constructs a leaf Unfold-operator. You must supply a sourceDB */
    public Unfold(DBManager localDB, DBManager sourceDB, String relation, 
                String pivot, String data) {
        super(localDB,LEAF); 
        if (!setSourceDB(sourceDB)) {
            System.out.println("Unfold operator wants to be a leaf.");
            System.exit(0);
        }
        relationName = relation;
        pivotName = pivot;
        dataName = data;
    }
    
    protected Vector computeLocalQueries() {
        Vector result =  new Vector(10);
        String query = new String();
        boolean firstTime = true;
        relSchema = sourceDB.getAttributeNames(relationName);
        // need to remove pivotName and dataName from relSchema since
        // that will be the remaining schema for the output relation.
        relSchema.removeElementAt(relSchema.indexOf(pivotName));
        relSchema.removeElementAt(relSchema.indexOf(dataName));
        
        // now we can flatten the vector out to a string since we'll
        // be making queries:
        String selectClause = new String();
        firstTime = true;
        for (Enumeration enum=relSchema.elements();
             enum.hasMoreElements();) {
            if (firstTime)
                firstTime = false;
            else {
                selectClause = selectClause + ", ";
            }
            selectClause += (String)enum.nextElement();
        }
        if (Main.DEBUG) System.out.println("UNFOLD: selectClause ="+selectClause);

        // now we retrieve all distinct values of the pivot attribute
        // and their counts
        pivotData = 
            sourceDB.executeViewQuery("select "
                                      + pivotName
                                      + ", count("
                                      + pivotName
                                      + ") from "
                                      + relationName
                                      + " group by "
                                      + pivotName);

        Vector pivotElements = pivotData.getTuples();
        // We had a bug here. We need to first fill the table with the
        // key values, then we add the data in. Otherwise, NULL values
        // are not properly handled. One side-effect is that the
        // output table will always be sorted (ascendingly) by its key.
        
        // we're creating a part of the select-clause containing nulls
        // for each new column except one and gives each column the
        // right alias name
        String nulls = new String();

        if (Main.DEBUG) System.out.println("UNFOLD: pivotElements ="+pivotElements.toString());

        if (pivotElements.size()>0) {
            nulls = "NULL as \""+(String)((Tuple)pivotElements.elementAt(0)).getVector().elementAt(0)+"\"";
            for (int i=1; i<pivotElements.size(); i++)
                nulls += ", NULL as \""+(String)((Tuple)pivotElements.elementAt(i)).getVector().elementAt(0)+"\"";
            
            query = "select distinct "
                + selectClause // = relSchema flattened into a string
                + ", "
                + nulls
                + " from "+relationName
                + " order by " 
                + selectClause;
        }
        else
            query = "select distinct "
                + selectClause
                + " from "+relationName
                +" order by "+selectClause;
        result.addElement(query);
        query = new String();
        // the table that is generated by this has the name
        // "outputTableName" which is a class variable of this' superclass

        // now we generate the rest as queries for single columns and
        // let the Table class do the rest


        String currentAttribute,pivotAtt;
        
        for (Enumeration enum=pivotElements.elements();
             enum.hasMoreElements();) {
            pivotAtt = (String)((Tuple)enum.nextElement()).getVector().elementAt(0);

            query += ("select "
                      +selectClause
                      +","
                      + dataName 
                      + " as "
                      + "\""+pivotAtt+"\""
                      + " from "
                      + relationName
                      + " where "
                      + pivotName
                      + " = '"+pivotAtt+"'"
                      + " order by " 
                      + selectClause
                      );

        result.addElement(query);
        query = new String();
        }

        if (Main.DEBUG) System.out.println("UNFOLD produced: "
                                           +result.toString());
        return result;
    }

    protected boolean executeQuery() {
        // outputTableName contains an intermediate table name that can be
        // used for output.
        Vector queries = computeLocalQueries();
        Table result;
        //  we have generate queries.  now we retrieve the maximum
        // display width of all data columns in the unfold query since
        // we need to override the column widths in the Table object.     
        int columnWidth =
            sourceDB.getMaximumDisplayWidth(relationName, dataName);

        // create a table object by executing the first query
        result = sourceDB.executeViewQuery((String)queries.elementAt(0));

        // table gets the intermediate name for this operator
        result.setRelName(outputTableName);


        // increase the column sizes to the maximum needed
        Integer aSize;
        Vector attSizes = result.getAttributeSizes();
        for (int i = 0 ; i < attSizes.size(); i++) {
            aSize = (Integer)attSizes.elementAt(i);
            if (aSize.intValue()==0) {
                aSize = new Integer(columnWidth); 
            }
            attSizes.setElementAt(aSize,i);
        }

        // now we can execute the remaining (update) queries and join
        // each new table with the existing one (in memory). We'll
        // teach the Table class to do that.

        // we don't need the first query anymore
        queries.removeElementAt(0);

        Table currentColumn;
        for (Enumeration enum = queries.elements();
             enum.hasMoreElements();
             ) {
            currentColumn = sourceDB.executeViewQuery
                ((String)enum.nextElement());
            // this overwrites values in one column and table must be
            // sorted by the same criterion. The data is passed as a
            // table with one attribute and the correct attribute name
            if (!result.setColumn(currentColumn))
                Util.exit("Error in UNFOLD node. Couldn't join new column");
        }
        return localDB.storeRelation(result);

    }

    /** propagate an update through this operator */
    /* package private */Vector propagateUpdate(Update upd) {
        //        if (Main.DEBUG) 
        //  System.err.println(this.toString()+": Trying to propagate: "+upd);
        // if update is not for me, return null

        Vector result = new Vector();
        String updPivot;
        if (!upd.relation.equalsIgnoreCase(relationName)) return null;

        if (upd instanceof DataUpdate) {
            DataUpdate updD = (DataUpdate)upd;
            // a data update
            if (updD.before==null) { // add-tuple
                    // PROBLEM (081000). The way I model data updates,
                    // I need to access the database to propagate
                    // updates. My update propagation right now
                    // assumes that a table knows its output table.
                    
                    // 081100: What the ****, we can just query that
                    // table, who cares...if it's too slow we can
                    // always just store the table in RAM

                updPivot =
                    (String)updD.after.getColumn(pivotName).elementAt(0);
                int pivotCount = findPivotCount(updPivot);
                if (pivotCount>0) {// A* doesn't change
                    incPivotCount(updPivot);
                }
                else { // A* changes
                    addPivot(updPivot);
                    result.addElement(new SchemaUpdate(outputTableName,updPivot,SchemaUpdate.ADD));
                }
                // then translate update query into vector of data
                // updates

                Table tmpTable = executeUpdateQuery(outputTableName, relSchema, updD.after);
                if (pivotCount==0) {// A* changes, change tmpTable. New column is last element!
                    ((Vector)tmpTable.getAttributeNames()).addElement(updPivot.toUpperCase());
                    // values should be changed, too, but it works like this.
                }
                String newValue = (String)updD.after.getColumn(dataName).elementAt(0);
                DataUpdate tmpUpdate;
                // now if this table has no rows, we insert the new
                // tuple. If it has one, we update that value, if
                // more, the table is weird.
                switch (tmpTable.getTuples().size()) {
                case 0:
                    // create add-tuple-update with the new schema
                    // HOW???
                    Table newTable = new Table(tmpTable);
                    newTable.setRelName("DUMMY");
                    Vector newAtts = newTable.getAttributeNames(); // wrong!
                    int columnCount = newAtts.size();
                    Vector newValues = new Vector(columnCount);
                    Vector updAttNames = updD.after.getAttributeNames();
                    Vector updValues  = ((Tuple)updD.after.getTuples().elementAt(0)).getVector();
                    int j;
                    for (int i = 0 ; i < columnCount; i++) {
                        j = updAttNames.indexOf(newAtts.elementAt(i));
                        if (j==-1)
                            newValues.insertElementAt(null,i);
                        else
                            newValues.insertElementAt(updValues.elementAt(j),i);
                    }
                    j = newAtts.indexOf(updPivot.toUpperCase());
                    newValues.setElementAt(newValue,j);
                    Tuple newTuple = new Tuple(newTable, newValues);
                    newTable.addTuple(newTuple);
                    tmpUpdate = new DataUpdate(outputTableName,null,new Table(newTable));
                    result.addElement(tmpUpdate);
                    break;
                case 1:
                    // an update-tuple-update
                    tmpUpdate = new
                        DataUpdate(outputTableName,new Table(tmpTable),new Table(tmpTable));
                    Vector tmpTupleVector =
                        ((Tuple)tmpUpdate.after.getTuples().elementAt(0)).getVector();
                    Vector tmpTupleAttributeNames = tmpUpdate.after.getAttributeNames();
                    // change the value to the new one
                    tmpTupleVector.setElementAt(newValue,tmpTupleAttributeNames.indexOf(updPivot.toUpperCase()));
                    result.addElement(tmpUpdate);
                    break;
                default:
                    Util.exit("Primary key violation... weird.");
                }

            }
            else if (updD.after==null) {// delete-tuple
                updPivot = 
                    (String)updD.before.getColumn(pivotName).elementAt(0);
                int pivotCount = findPivotCount(updPivot);
                if (pivotCount==1) { // delete that attribute
                    delPivot(updPivot);
                    SchemaUpdate tmpUpdate = new SchemaUpdate(outputTableName, updPivot.toUpperCase(), SchemaUpdate.DELETE);
                    result.addElement(tmpUpdate);
                }
                else if (pivotCount>1) { // update. Set corresponding value to null
                    decPivotCount(updPivot);
                    Table tmpTable = executeUpdateQuery(outputTableName, relSchema, updD.before);
                    if (tmpTable==null)
                        Util.exit("Invalid update. Tuple to be deleted doesn't exist.");
 
                   DataUpdate tmpUpdate = new
                        DataUpdate(outputTableName,new Table(tmpTable),new Table(tmpTable));
                    Vector tmpTupleVector =
                        ((Tuple)tmpUpdate.after.getTuples().elementAt(0)).getVector();
                    Vector tmpTupleAttributeNames = tmpUpdate.after.getAttributeNames();
                    // change the value to null
                    tmpTupleVector.setElementAt(null,tmpTupleAttributeNames.indexOf(updPivot.toUpperCase()));
                    result.addElement(tmpUpdate);
                    
                }
                else // invalid delete-tuple
                     Util.exit("Invalid update. Tuple to be deleted doesn't exist.");
            }
            else { // update-tuple
                if (Main.DEBUG) System.out.println("Entering recursive call becuase of update-tuple update");
                // recursive call. Make sure we don't have update-tuple-changes here!!!!
                // add-tuple
                result = this.propagateUpdate(new DataUpdate(updD.relation,null,updD.after));
                // delete-tuple
                result = QENode.mergeVectors(result,this.propagateUpdate(new DataUpdate(updD.relation,updD.before,null)));
            }
        }
        else {
            SchemaUpdate updS = (SchemaUpdate)upd;
            SchemaUpdate tmpSchemaUpdate;
            // a schema change
            if (updS.attribute!=null) {
                // an attribute update
                switch (updS.type) {
                case SchemaUpdate.ADD:
                    tmpSchemaUpdate = new SchemaUpdate(updS);
                    tmpSchemaUpdate.relation = new String(outputTableName);
                    result.addElement(tmpSchemaUpdate);
                    // maintain relation schema vector
                    relSchema.addElement(updS.attribute);
                    break;
                case SchemaUpdate.DELETE:
                    if (updS.attribute.equalsIgnoreCase(pivotName) ||
                        updS.attribute.equalsIgnoreCase(dataName)) {
                        Util.invalidView();
                    }
                    else
                        // for the time being (which is probably
                        // forever), we won't propagate here since we
                        // don't know if there is a star in the select-clause
                        /* result.addElement(new SchemaUpdate(updS));
                         * */
                        Util.invalidView();
                    break;
                case SchemaUpdate.RENAME:
                    if (updS.attribute.equalsIgnoreCase(pivotName)) {
                        pivotName = updS.newName;
                    } 
                    else if
                        (updS.attribute.equalsIgnoreCase(dataName)) {
                        dataName = updS.newName;
                    }
                    else {
                    tmpSchemaUpdate = new SchemaUpdate(updS);
                    tmpSchemaUpdate.relation = new String(outputTableName);
                    result.addElement(tmpSchemaUpdate);
                    // maintain relSchema
                    relSchema.setElementAt(updS.newName,relSchema.indexOf(updS.attribute));
                    }
                }
            }
            else {
                switch (updS.type) {
                case SchemaUpdate.ADD:
                    // do nothing... makes no sense
                    break;
                case SchemaUpdate.DELETE:
                    tmpSchemaUpdate = new SchemaUpdate(updS);
                    tmpSchemaUpdate.relation = new String(outputTableName);
                    result.addElement(tmpSchemaUpdate);
                    break;
                case SchemaUpdate.RENAME:
                    relationName = updS.newName;
                    break;
                }
            }
        }

        // Here, we should find tuples that have nulls in all of their
        // unfolded columns, and delete them. I don't feel like it
        // right now... (100600)

        //this:
        /*
        Vector unfoldedAttNames = pivotData.getColumn(0);
        String nullQuery = "select * from "+outputTableName+" where ";
        for (int i = 0  ; i < unfoldedAttNames.size(); i++) {
            nullQuery += unfoldedAttNames.elementAt(i)+" is null ";
            if (i<unfoldedAttNames.size()-1) nullQuery += "and ";
        }
        
        Table allNulls = localDB.executeViewQuery(nullQuery);
        result = QENode.mergeVectors(result,tableToDeleteUpdates(outputTableName,allNulls));
        */
        // doesn't work since we have schema changes befroe this is ever executed

        if (Main.DEBUG) System.out.println("Incremental Update Propagation. UNFOLD gave us: "+result.toString());

        // the QENode above this must execute these updates!!!
        return result;
    }

}

