//
// Andreas Koeller July 2000
//

package schemaSQLParser.visitor;
import schemaSQLParser.syntaxtree.*;
import java.util.*;
import treeBuilder.*;
import algebraTree.Util;

/**
 * SchemaSQL Visitor. Generates algebra tree from a parsed SchemaSQL query.
 */

public class SchemaSQLVisitor extends DepthFirstVisitor {

    /** This visitor creates a Builder object and fills it out.
     */
    
    public Builder bld;

    public SchemaSQLVisitor(Builder bldP) {
        bld = bldP;
    }
    //
    // Auto class visitors--probably don't need to be overridden.
    //
    public void visit(NodeList n) {
        for ( Enumeration e = n.elements(); e.hasMoreElements(); )
            ((Node)e.nextElement()).accept(this);
    }

    public void visit(NodeListOptional n) {
        if ( n.present() )
            for ( Enumeration e = n.elements(); e.hasMoreElements(); )
                ((Node)e.nextElement()).accept(this);
    }

    public void visit(NodeOptional n) {
        if ( n.present() )
            n.node.accept(this);
    }

    public void visit(NodeSequence n) {
        for ( Enumeration e = n.elements(); e.hasMoreElements(); )
            ((Node)e.nextElement()).accept(this);
    }
 
    public void visit(NodeToken n) { }

    //
    // User-generated visitor methods below
    //

    //============================================================
    //============================================================
    //============================================================
    //============================================================
    /** global variables for the visitor */
    int fromItemIndex = -1;
    tbFromItem curFromItem;
    String curObjectName;
    String curTableReference;
    String curAttributeReference;
    String curConstant;
    String curViewItem;
    String curColumnAlias;
    tbSelectItem curSelectItem;
    int curRelop;
    tbConstraint curConstraint;

    //============================================================
    //============================================================
    //============================================================
    //============================================================


    /**
     * nodeChoice -> ProcedureDeclaration()
     *       | FunctionDeclaration()
     *       | SequenceOfStatements()
     */
    public void visit(CompilationUnit n) {
       
        n.nodeChoice.accept(this);
    }

    /**
     * nodeToken -> "DECLARE"
     * declarations -> Declarations()
     */
    public void visit(DeclarationSection n) {
        n.nodeToken.accept(this);
        n.declarations.accept(this);
    }

    /**
     * nodeList -> ( ( IdentifierDeclaration() | CursorDeclaration() | PragmaDeclaration() ) ";" )+
     * nodeListOptional -> ( ProcedureDeclaration() | FunctionDeclaration() )*
     */
    public void visit(Declarations n) {
        n.nodeList.accept(this);
        n.nodeListOptional.accept(this);
    }

    /**
     * nodeToken -> <S_IDENTIFIER>
     * nodeChoice -> ( ConstantDeclaration() | VariableDeclaration() | ExceptionDeclaration() )
     */
    public void visit(IdentifierDeclaration n) {
        n.nodeToken.accept(this);
        n.nodeChoice.accept(this);
    }

    /**
     * nodeToken -> "CURSOR"
     * nodeToken1 -> <S_IDENTIFIER>
     * nodeOptional -> [ "(" ParameterList() ")" ]
     * nodeToken2 -> "IS"
     * selectStatement -> SelectStatement()
     */
    public void visit(CursorDeclaration n) {
        n.nodeToken.accept(this);
        n.nodeToken1.accept(this);
        n.nodeOptional.accept(this);
        n.nodeToken2.accept(this);
        n.selectStatement.accept(this);
    }

    /**
     * nodeToken -> "PROCEDURE"
     * nodeToken1 -> <S_IDENTIFIER>
     * nodeOptional -> [ "(" ParameterList() ")" ]
     * nodeChoice -> ( ";" | "IS" ProcedureBody() )
     */
    public void visit(ProcedureDeclaration n) {
        n.nodeToken.accept(this);
        n.nodeToken1.accept(this);
        n.nodeOptional.accept(this);
        n.nodeChoice.accept(this);
    }

    /**
     * nodeOptional -> [ Declarations() ]
     * beginEndBlock -> BeginEndBlock()
     */
    public void visit(ProcedureBody n) {
        n.nodeOptional.accept(this);
        n.beginEndBlock.accept(this);
    }

    /**
     * nodeToken -> "FUNCTION"
     * nodeToken1 -> <S_IDENTIFIER>
     * nodeOptional -> [ "(" ParameterList() ")" ]
     * nodeToken2 -> "RETURN"
     * typeDeclaration -> TypeDeclaration()
     * nodeChoice -> ( ";" | "IS" FunctionBody() )
     */
    public void visit(FunctionDeclaration n) {
        n.nodeToken.accept(this);
        n.nodeToken1.accept(this);
        n.nodeOptional.accept(this);
        n.nodeToken2.accept(this);
        n.typeDeclaration.accept(this);
        n.nodeChoice.accept(this);
    }

    /**
     * nodeOptional -> [ Declarations() ]
     * beginEndBlock -> BeginEndBlock()
     */
    public void visit(FunctionBody n) {
        n.nodeOptional.accept(this);
        n.beginEndBlock.accept(this);
    }

    /**
     * nodeToken -> "PRAGMA"
     * nodeToken1 -> "EXCEPTION_INIT"
     * nodeToken2 -> "("
     * numOrID -> NumOrID()
     * nodeToken3 -> ","
     * numOrID1 -> NumOrID()
     * nodeToken4 -> ")"
     */
    public void visit(PragmaDeclaration n) {
        n.nodeToken.accept(this);
        n.nodeToken1.accept(this);
        n.nodeToken2.accept(this);
        n.numOrID.accept(this);
        n.nodeToken3.accept(this);
        n.numOrID1.accept(this);
        n.nodeToken4.accept(this);
    }

    /**
     * typeDeclaration -> TypeDeclaration()
     * nodeOptional -> [ "NOT" "NULL" ]
     * nodeOptional1 -> [ ( ":=" | "DEFAULT" ) PlSqlExpression() ]
     */
    public void visit(VariableDeclaration n) {
        n.typeDeclaration.accept(this);
        n.nodeOptional.accept(this);
        n.nodeOptional1.accept(this);
    }

    /**
     * nodeToken -> "CONSTANT"
     * typeDeclaration -> TypeDeclaration()
     * nodeOptional -> [ "NOT" "NULL" ]
     * nodeChoice -> ( ":=" | "DEFAULT" )
     * plSqlExpression -> PlSqlExpression()
     */
    public void visit(ConstantDeclaration n) {
        n.nodeToken.accept(this);
        n.typeDeclaration.accept(this);
        n.nodeOptional.accept(this);
        n.nodeChoice.accept(this);
        n.plSqlExpression.accept(this);
    }

    /**
     * nodeChoice -> BasicDataTypeDeclaration()
     *       | ( <S_IDENTIFIER> ( "%TYPE" | "%ROWTYPE" ) )
     *       | TableColumn() "%TYPE"
     *       | <S_IDENTIFIER>
     */
    public void visit(TypeDeclaration n) {
        n.nodeChoice.accept(this);
    }

    /**
     * nodeChoice -> ( "CHAR" | "VARCHAR" | "VARCHAR2" | "INTEGER" | "NUMBER" | "NATURAL" | "REAL" | "FLOAT" ) [ "(" <S_NUMBER> [ "," <S_NUMBER> ] ")" ]
     *       | "DATE"
     *       | "BINARY_INTEGER"
     *       | "BOOLEAN"
     */
    public void visit(BasicDataTypeDeclaration n) {
        n.nodeChoice.accept(this);
    }

    /**
     * nodeToken -> "EXCEPTION"
     */
    public void visit(ExceptionDeclaration n) {
        n.nodeToken.accept(this);
    }

    /**
     * nodeToken -> "BEGIN"
     * sequenceOfStatements -> SequenceOfStatements()
     * nodeOptional -> [ ExceptionBlock() ]
     * nodeToken1 -> "END"
     * nodeOptional1 -> [ <S_IDENTIFIER> ]
     * nodeToken2 -> ";"
     */
    public void visit(BeginEndBlock n) {
        n.nodeToken.accept(this);
        n.sequenceOfStatements.accept(this);
        n.nodeOptional.accept(this);
        n.nodeToken1.accept(this);
        n.nodeOptional1.accept(this);
        n.nodeToken2.accept(this);
    }

    /**
     * nodeList -> ( PLSQLStatement() )+
     */
    public void visit(SequenceOfStatements n) {
        n.nodeList.accept(this);
    }

    /**
     * nodeToken -> "EXCEPTION"
     * nodeList -> ( ExceptionHandler() )+
     */
    public void visit(ExceptionBlock n) {
        n.nodeToken.accept(this);
        n.nodeList.accept(this);
    }

    /**
     * nodeToken -> "WHEN"
     * nodeChoice -> ( <S_IDENTIFIER> ( "OR" <S_IDENTIFIER> )* | "OTHERS" )
     * nodeToken1 -> "THEN"
     * sequenceOfStatements -> SequenceOfStatements()
     */
    public void visit(ExceptionHandler n) {
        n.nodeToken.accept(this);
        n.nodeChoice.accept(this);
        n.nodeToken1.accept(this);
        n.sequenceOfStatements.accept(this);
    }

    /**
     * nodeChoice -> SubroutineCall()
     *       | AssignmentStatement()
     *       | ExitStatement()
     *       | GotoStatement()
     *       | IfStatement()
     *       | LoopStatement()
     *       | NullStatement()
     *       | RaiseStatement()
     *       | ReturnStatement()
     *       | SQLStatement()
     *       | LabelDeclaration()
     *       | [ DeclarationSection() ] BeginEndBlock()
     */
    public void visit(PLSQLStatement n) {
        n.nodeChoice.accept(this);
    }

    /**
     * nodeToken -> "<<"
     * nodeToken1 -> <S_IDENTIFIER>
     * nodeToken2 -> ">>"
     */
    public void visit(LabelDeclaration n) {
        n.nodeToken.accept(this);
        n.nodeToken1.accept(this);
        n.nodeToken2.accept(this);
    }

    /**
     * nodeChoice -> CloseStatement()
     *       | CommitStatement()
     *       | DeleteStatement()
     *       | FetchStatement()
     *       | InsertStatement()
     *       | LockTableStatement()
     *       | OpenStatement()
     *       | RollbackStatement()
     *       | SavepointStatement()
     *       | QueryStatement()
     *       | SetTransactionStatement()
     *       | UpdateStatement()
     */
    public void visit(SQLStatement n) {
        n.nodeChoice.accept(this);
    }

    /**
     * nodeToken -> <S_IDENTIFIER>
     * nodeOptional -> [ "." <S_IDENTIFIER> ]
     * nodeOptional1 -> [ "(" Arguments() ")" ]
     * nodeToken1 -> ";"
     */
    public void visit(SubroutineCall n) {
        n.nodeToken.accept(this);
        n.nodeOptional.accept(this);
        n.nodeOptional1.accept(this);
        n.nodeToken1.accept(this);
    }

    /**
     * nodeChoice -> ( ( <S_IDENTIFIER> [ "." <S_IDENTIFIER> ] ":=" PlSqlExpression() ) | ( <S_BIND> ":=" PlSqlExpression() ) )
     * nodeToken -> ";"
     */
    public void visit(AssignmentStatement n) {
        n.nodeChoice.accept(this);
        n.nodeToken.accept(this);
    }

    /**
     * nodeToken -> "EXIT"
     * nodeOptional -> [ <S_IDENTIFIER> ]
     * nodeOptional1 -> [ "WHEN" PlSqlExpression() ]
     * nodeToken1 -> ";"
     */
    public void visit(ExitStatement n) {
        n.nodeToken.accept(this);
        n.nodeOptional.accept(this);
        n.nodeOptional1.accept(this);
        n.nodeToken1.accept(this);
    }

    /**
     * nodeToken -> "GOTO"
     * nodeToken1 -> <S_IDENTIFIER>
     * nodeToken2 -> ";"
     */
    public void visit(GotoStatement n) {
        n.nodeToken.accept(this);
        n.nodeToken1.accept(this);
        n.nodeToken2.accept(this);
    }

    /**
     * nodeToken -> "IF"
     * plSqlExpression -> PlSqlExpression()
     * nodeToken1 -> "THEN"
     * sequenceOfStatements -> SequenceOfStatements()
     * nodeListOptional -> ( "ELSIF" PlSqlExpression() "THEN" SequenceOfStatements() )*
     * nodeOptional -> [ "ELSE" SequenceOfStatements() ]
     * nodeToken2 -> "END"
     * nodeToken3 -> "IF"
     * nodeOptional1 -> [ <S_IDENTIFIER> ]
     * nodeToken4 -> ";"
     */
    public void visit(IfStatement n) {
        n.nodeToken.accept(this);
        n.plSqlExpression.accept(this);
        n.nodeToken1.accept(this);
        n.sequenceOfStatements.accept(this);
        n.nodeListOptional.accept(this);
        n.nodeOptional.accept(this);
        n.nodeToken2.accept(this);
        n.nodeToken3.accept(this);
        n.nodeOptional1.accept(this);
        n.nodeToken4.accept(this);
    }

    /**
     * nodeChoice -> NormalLoop()
     *       | WhileLoop()
     *       | ForLoop()
     */
    public void visit(LoopStatement n) {
        n.nodeChoice.accept(this);
    }

    /**
     * nodeToken -> "LOOP"
     * sequenceOfStatements -> SequenceOfStatements()
     * nodeToken1 -> "END"
     * nodeToken2 -> "LOOP"
     * nodeOptional -> [ <S_IDENTIFIER> ]
     * nodeToken3 -> ";"
     */
    public void visit(NormalLoop n) {
        n.nodeToken.accept(this);
        n.sequenceOfStatements.accept(this);
        n.nodeToken1.accept(this);
        n.nodeToken2.accept(this);
        n.nodeOptional.accept(this);
        n.nodeToken3.accept(this);
    }

    /**
     * nodeToken -> "WHILE"
     * plSqlExpression -> PlSqlExpression()
     * normalLoop -> NormalLoop()
     */
    public void visit(WhileLoop n) {
        n.nodeToken.accept(this);
        n.plSqlExpression.accept(this);
        n.normalLoop.accept(this);
    }

    /**
     * nodeChoice -> NumericForLoop()
     *       | CursorForLoop()
     */
    public void visit(ForLoop n) {
        n.nodeChoice.accept(this);
    }

    /**
     * nodeToken -> "FOR"
     * nodeToken1 -> <S_IDENTIFIER>
     * nodeToken2 -> "IN"
     * nodeOptional -> [ "REVERSE" ]
     * plSqlSimpleExpression -> PlSqlSimpleExpression()
     * nodeToken3 -> ".."
     */
    public void visit(NumericForLoopLookahead n) {
        n.nodeToken.accept(this);
        n.nodeToken1.accept(this);
        n.nodeToken2.accept(this);
        n.nodeOptional.accept(this);
        n.plSqlSimpleExpression.accept(this);
        n.nodeToken3.accept(this);
    }

    /**
     * nodeToken -> "FOR"
     * nodeToken1 -> <S_IDENTIFIER>
     * nodeToken2 -> "IN"
     * nodeOptional -> [ "REVERSE" ]
     * plSqlSimpleExpression -> PlSqlSimpleExpression()
     * nodeToken3 -> ".."
     * plSqlSimpleExpression1 -> PlSqlSimpleExpression()
     * normalLoop -> NormalLoop()
     */
    public void visit(NumericForLoop n) {
        n.nodeToken.accept(this);
        n.nodeToken1.accept(this);
        n.nodeToken2.accept(this);
        n.nodeOptional.accept(this);
        n.plSqlSimpleExpression.accept(this);
        n.nodeToken3.accept(this);
        n.plSqlSimpleExpression1.accept(this);
        n.normalLoop.accept(this);
    }

    /**
     * nodeToken -> "FOR"
     * nodeToken1 -> <S_IDENTIFIER>
     * nodeToken2 -> "IN"
     * nodeChoice -> ( <S_IDENTIFIER> [ "(" Arguments() ")" ] | "(" SelectStatement() ")" )
     * normalLoop -> NormalLoop()
     */
    public void visit(CursorForLoop n) {
        n.nodeToken.accept(this);
        n.nodeToken1.accept(this);
        n.nodeToken2.accept(this);
        n.nodeChoice.accept(this);
        n.normalLoop.accept(this);
    }

    /**
     * nodeToken -> "NULL"
     * nodeToken1 -> ";"
     */
    public void visit(NullStatement n) {
        n.nodeToken.accept(this);
        n.nodeToken1.accept(this);
    }

    /**
     * nodeToken -> "RAISE"
     * nodeOptional -> [ <S_IDENTIFIER> ]
     * nodeToken1 -> ";"
     */
    public void visit(RaiseStatement n) {
        n.nodeToken.accept(this);
        n.nodeOptional.accept(this);
        n.nodeToken1.accept(this);
    }

    /**
     * nodeToken -> "RETURN"
     * nodeOptional -> [ PlSqlExpression() ]
     * nodeToken1 -> ";"
     */
    public void visit(ReturnStatement n) {
        n.nodeToken.accept(this);
        n.nodeOptional.accept(this);
        n.nodeToken1.accept(this);
    }

    /**
     * nodeToken -> "CLOSE"
     * nodeToken1 -> <S_IDENTIFIER>
     * nodeToken2 -> ";"
     */
    public void visit(CloseStatement n) {
        n.nodeToken.accept(this);
        n.nodeToken1.accept(this);
        n.nodeToken2.accept(this);
    }

    /**
     * nodeToken -> "COMMIT"
     * nodeOptional -> [ "WORK" ]
     * nodeOptional1 -> [ "COMMENT" <S_CHAR_LITERAL> ]
     * nodeToken1 -> ";"
     */
    public void visit(CommitStatement n) {
        n.nodeToken.accept(this);
        n.nodeOptional.accept(this);
        n.nodeOptional1.accept(this);
        n.nodeToken1.accept(this);
    }

    /**
     * nodeToken -> "FETCH"
     * nodeSequence -> ( <S_IDENTIFIER> )
     * nodeToken1 -> "INTO"
     * nodeChoice -> ( <S_IDENTIFIER> | <S_BIND> )
     * nodeListOptional -> ( "," ( <S_IDENTIFIER> | <S_BIND> ) )*
     * nodeToken2 -> ";"
     */
    public void visit(FetchStatement n) {
        n.nodeToken.accept(this);
        n.nodeSequence.accept(this);
        n.nodeToken1.accept(this);
        n.nodeChoice.accept(this);
        n.nodeListOptional.accept(this);
        n.nodeToken2.accept(this);
    }

    /**
     * nodeToken -> "LOCK"
     * nodeToken1 -> "TABLE"
     * tableReference -> TableReference()
     * nodeListOptional -> ( "," TableReference() )*
     * nodeToken2 -> "IN"
     * lockMode -> LockMode()
     * nodeToken3 -> "MODE"
     * nodeOptional -> [ "NOWAIT" ]
     * nodeToken4 -> ";"
     */
    public void visit(LockTableStatement n) {
        n.nodeToken.accept(this);
        n.nodeToken1.accept(this);
        n.tableReference.accept(this);
        n.nodeListOptional.accept(this);
        n.nodeToken2.accept(this);
        n.lockMode.accept(this);
        n.nodeToken3.accept(this);
        n.nodeOptional.accept(this);
        n.nodeToken4.accept(this);
    }

    /**
     * nodeToken -> "OPEN"
     * nodeToken1 -> <S_IDENTIFIER>
     * nodeOptional -> [ "(" Arguments() ")" ]
     * nodeToken2 -> ";"
     */
    public void visit(OpenStatement n) {
        n.nodeToken.accept(this);
        n.nodeToken1.accept(this);
        n.nodeOptional.accept(this);
        n.nodeToken2.accept(this);
    }

    /**
     * nodeToken -> "ROLLBACK"
     * nodeOptional -> [ "WORK" ]
     * nodeOptional1 -> [ "TO" [ "SAVEPOINT" ] <S_IDENTIFIER> ]
     * nodeOptional2 -> [ "COMMENT" <S_CHAR_LITERAL> ]
     * nodeToken1 -> ";"
     */
    public void visit(RollbackStatement n) {
        n.nodeToken.accept(this);
        n.nodeOptional.accept(this);
        n.nodeOptional1.accept(this);
        n.nodeOptional2.accept(this);
        n.nodeToken1.accept(this);
    }

    /**
     * nodeToken -> "SET"
     * nodeToken1 -> "TRANSACTION"
     * nodeChoice -> ( ( "READ" ( "ONLY" | "WRITE" ) ) | ( "USE" "ROLLBACK" "SEGMENT" <S_IDENTIFIER> ) )
     * nodeToken2 -> ";"
     */
    public void visit(SetTransactionStatement n) {
        n.nodeToken.accept(this);
        n.nodeToken1.accept(this);
        n.nodeChoice.accept(this);
        n.nodeToken2.accept(this);
    }

    /**
     * nodeChoice -> ( "ROW" ( "SHARE" | "EXCLUSIVE" ) )
     *       | ( "SHARE" [ "UPDATE" | ( "ROW" "EXCLUSIVE" ) ] )
     *       | ( "EXCLUSIVE" )
     */
    public void visit(LockMode n) {
        n.nodeChoice.accept(this);
    }

    /**
     * nodeToken -> "SAVEPOINT"
     * nodeToken1 -> <S_IDENTIFIER>
     * nodeToken2 -> ";"
     */
    public void visit(SavepointStatement n) {
        n.nodeToken.accept(this);
        n.nodeToken1.accept(this);
        n.nodeToken2.accept(this);
    }

    /**
     * nodeToken -> "UPDATE"
     * tableReference -> TableReference()
     * nodeOptional -> [ <S_IDENTIFIER> ]
     * nodeToken1 -> "SET"
     * columnValues -> ColumnValues()
     * nodeOptional1 -> [ "WHERE" ( SQLExpression() | "CURRENT" "OF" <S_IDENTIFIER> ) ]
     * nodeToken2 -> ";"
     */
    public void visit(UpdateStatement n) {
        n.nodeToken.accept(this);
        n.tableReference.accept(this);
        n.nodeOptional.accept(this);
        n.nodeToken1.accept(this);
        n.columnValues.accept(this);
        n.nodeOptional1.accept(this);
        n.nodeToken2.accept(this);
    }

    /**
     * tableColumn -> TableColumn()
     * nodeToken -> "="
     * updatedValue -> UpdatedValue()
     * nodeListOptional -> ( "," TableColumn() "=" UpdatedValue() )*
     */
    public void visit(ColumnValues n) {
        n.tableColumn.accept(this);
        n.nodeToken.accept(this);
        n.updatedValue.accept(this);
        n.nodeListOptional.accept(this);
    }

    /**
     * nodeChoice -> "(" SelectStatement() ")"
     *       | PlSqlExpression()
     */
    public void visit(UpdatedValue n) {
        n.nodeChoice.accept(this);
    }

    /**
     * nodeToken -> "INSERT"
     * nodeToken1 -> "INTO"
     * tableReference -> TableReference()
     * nodeOptional -> [ "(" TableColumn() ( "," TableColumn() )* ")" ]
     * nodeChoice -> ( "VALUES" "(" PlSqlExpressionList() ")" | SelectStatement() )
     * nodeToken2 -> ";"
     */
    public void visit(InsertStatement n) {
        n.nodeToken.accept(this);
        n.nodeToken1.accept(this);
        n.tableReference.accept(this);
        n.nodeOptional.accept(this);
        n.nodeChoice.accept(this);
        n.nodeToken2.accept(this);
    }

    /**
     * nodeToken -> "DELETE"
     * nodeOptional -> [ "FROM" ]
     * tableReference -> TableReference()
     * nodeOptional1 -> [ <S_IDENTIFIER> ]
     * nodeOptional2 -> [ "WHERE" ( SQLExpression() | "CURRENT" "OF" <S_IDENTIFIER> ) ]
     * nodeToken1 -> ";"
     */
    public void visit(DeleteStatement n) {
        n.nodeToken.accept(this);
        n.nodeOptional.accept(this);
        n.tableReference.accept(this);
        n.nodeOptional1.accept(this);
        n.nodeOptional2.accept(this);
        n.nodeToken1.accept(this);
    }

    /**
     * nodeOptional -> [ CreateViewClause() ]
     * selectStatement -> SelectStatement()
     * nodeToken -> ";"
     */
    public void visit(QueryStatement n) {
        n.nodeOptional.accept(this);
        n.selectStatement.accept(this);
        n.nodeToken.accept(this);
    }

    /**
     * nodeToken -> "CREATE"
     * nodeToken1 -> "VIEW"
     * viewReference -> ViewReference()
     * createViewList -> CreateViewList()
     * nodeToken2 -> "AS"
     */
    public void visit(CreateViewClause n) {
        
        n.nodeToken.accept(this);
        n.nodeToken1.accept(this);
        n.viewReference.accept(this);
        n.createViewList.accept(this);
        n.nodeToken2.accept(this);
    }

    /**
     * plSqlAndExpression -> PlSqlAndExpression()
     * nodeListOptional -> ( "OR" PlSqlAndExpression() )*
     */
    public void visit(PlSqlExpression n) {
        n.plSqlAndExpression.accept(this);
        n.nodeListOptional.accept(this);
    }

    /**
     * plSqlUnaryLogicalExpression -> PlSqlUnaryLogicalExpression()
     * nodeListOptional -> ( "AND" PlSqlUnaryLogicalExpression() )*
     */
    public void visit(PlSqlAndExpression n) {
        n.plSqlUnaryLogicalExpression.accept(this);
        n.nodeListOptional.accept(this);
    }

    /**
     * nodeOptional -> [ "NOT" ]
     * plSqlRelationalExpression -> PlSqlRelationalExpression()
     */
    public void visit(PlSqlUnaryLogicalExpression n) {
        n.nodeOptional.accept(this);
        n.plSqlRelationalExpression.accept(this);
    }

    /**
     * plSqlSimpleExpression -> PlSqlSimpleExpression()
     * nodeOptional -> ( Relop() PlSqlSimpleExpression() | PlSqlInClause() | PlSqlBetweenClause() | PlSqlLikeClause() | IsNullClause() )?
     */
    public void visit(PlSqlRelationalExpression n) {
        n.plSqlSimpleExpression.accept(this);
        n.nodeOptional.accept(this);
    }

    /**
     * plSqlExpression -> PlSqlExpression()
     * nodeListOptional -> ( "," PlSqlExpression() )*
     */
    public void visit(PlSqlExpressionList n) {
        n.plSqlExpression.accept(this);
        n.nodeListOptional.accept(this);
    }

    /**
     * nodeOptional -> [ "NOT" ]
     * nodeToken -> "IN"
     * nodeToken1 -> "("
     * plSqlExpressionList -> PlSqlExpressionList()
     * nodeToken2 -> ")"
     */
    public void visit(PlSqlInClause n) {
        n.nodeOptional.accept(this);
        n.nodeToken.accept(this);
        n.nodeToken1.accept(this);
        n.plSqlExpressionList.accept(this);
        n.nodeToken2.accept(this);
    }

    /**
     * nodeOptional -> [ "NOT" ]
     * nodeToken -> "BETWEEN"
     * plSqlSimpleExpression -> PlSqlSimpleExpression()
     * nodeToken1 -> "AND"
     * plSqlSimpleExpression1 -> PlSqlSimpleExpression()
     */
    public void visit(PlSqlBetweenClause n) {
        n.nodeOptional.accept(this);
        n.nodeToken.accept(this);
        n.plSqlSimpleExpression.accept(this);
        n.nodeToken1.accept(this);
        n.plSqlSimpleExpression1.accept(this);
    }

    /**
     * nodeOptional -> [ "NOT" ]
     * nodeToken -> "LIKE"
     * plSqlSimpleExpression -> PlSqlSimpleExpression()
     */
    public void visit(PlSqlLikeClause n) {
        n.nodeOptional.accept(this);
        n.nodeToken.accept(this);
        n.plSqlSimpleExpression.accept(this);
    }

    /**
     * nodeToken -> "IS"
     * nodeOptional -> [ "NOT" ]
     * nodeToken1 -> "NULL"
     */
    public void visit(IsNullClause n) {
        n.nodeToken.accept(this);
        n.nodeOptional.accept(this);
        n.nodeToken1.accept(this);
    }

    /**
     * plSqlMultiplicativeExpression -> PlSqlMultiplicativeExpression()
     * nodeListOptional -> ( ( "+" | "-" | "||" ) PlSqlMultiplicativeExpression() )*
     */
    public void visit(PlSqlSimpleExpression n) {
        n.plSqlMultiplicativeExpression.accept(this);
        n.nodeListOptional.accept(this);
    }

    /**
     * plSqlExpotentExpression -> PlSqlExpotentExpression()
     * nodeListOptional -> ( ( "*" | "/" ) PlSqlExpotentExpression() )*
     */
    public void visit(PlSqlMultiplicativeExpression n) {
        n.plSqlExpotentExpression.accept(this);
        n.nodeListOptional.accept(this);
    }

    /**
     * plSqlUnaryExpression -> PlSqlUnaryExpression()
     * nodeListOptional -> ( "**" PlSqlUnaryExpression() )*
     */
    public void visit(PlSqlExpotentExpression n) {
        n.plSqlUnaryExpression.accept(this);
        n.nodeListOptional.accept(this);
    }

    /**
     * nodeChoice -> ( ( "+" | "-" ) PlSqlPrimaryExpression() )
     *       | PlSqlPrimaryExpression()
     */
    public void visit(PlSqlUnaryExpression n) {
        n.nodeChoice.accept(this);
    }

    /**
     * nodeChoice -> "NULL"
     *       | ( <S_IDENTIFIER> ( "%FOUND" | "%NOTFOUND" | "%ISOPEN" | "%ROWCOUNT" ) )
     *       | ( <S_IDENTIFIER> "(" Arguments() ")" )
     *       | ( <S_IDENTIFIER> [ "." <S_IDENTIFIER> ] )
     *       | ( "SQL" ( "%FOUND" | "%NOTFOUND" | "%ISOPEN" | "%ROWCOUNT" ) )
     *       | <S_NUMBER>
     *       | <S_CHAR_LITERAL>
     *       | <S_BIND>
     *       | "(" PlSqlExpression() ")"
     */
    public void visit(PlSqlPrimaryExpression n) {
        n.nodeChoice.accept(this);
    }

    /**
     * oracleObjectName -> OracleObjectName()
     * nodeOptional -> [ "." OracleObjectName() ]
     */
    public void visit(TableColumn n) {
        n.oracleObjectName.accept(this);
        curTableReference = new String(curObjectName); 
        n.nodeOptional.accept(this);
        if (n.nodeOptional.present())
            curAttributeReference = new String(curObjectName);
        else
            curAttributeReference =null;
    }

    /**
     * nodeChoice -> <S_QUOTED_IDENTIFIER>
     *       | <S_IDENTIFIER>
     */
    public void visit(OracleObjectName n) {
        //System.out.println("OracleObjectName: which:"+n.nodeChoice.choice.toString());
        if (n.getParent() instanceof SelectItem && curObjectName!=null) {
            Util.exit("Expressions of type a.b.* in select-clause not handled");
            
        }
            
        // always upper case!!!
        curObjectName = n.nodeChoice.choice.toString().toUpperCase();
        if (n.nodeChoice.which == 0) // S_QUOTED_IDENTIFIER
            curObjectName = 
                curObjectName.substring(1,curObjectName.length()-1);
        n.nodeChoice.accept(this);
    }

    /**
     * nodeChoice -> "="
     *       | "!="
     *       | "#"
     *       | "<>"
     *       | ">"
     *       | ">="
     *       | "<"
     *       | "<="
     */
    public void visit(Relop n) {
        String tmp;
        n.nodeChoice.accept(this);
        tmp = n.nodeChoice.choice.toString();
        if (tmp.equals("="))
            curRelop = tbConstraint.EQ;
        else 
        if (tmp.equals("!="))
            curRelop = tbConstraint.NE;
        else 
        if (tmp.equals("<>"))
            curRelop = tbConstraint.LG;
        else 
        if (tmp.equals(">"))
            curRelop = tbConstraint.GT;
        else 
        if (tmp.equals(">="))
            curRelop = tbConstraint.GE;
        else 
        if (tmp.equals("<"))
            curRelop = tbConstraint.LT;
        else 
        if (tmp.equals("<="))
            curRelop = tbConstraint.LE;
        else 
            curRelop = tbConstraint.ERROR;
    }

    /**
     * oracleObjectName -> OracleObjectName()
     * nodeOptional -> [ "." OracleObjectName() ]
     */
    public void visit(TableReference n) {
        String myname;
        n.oracleObjectName.accept(this);

        myname = new String(curObjectName);
        
        n.nodeOptional.accept(this);

        if (n.getParent() instanceof FromItem) {
            if (n.nodeOptional.present()) {
                curFromItem.type = Builder.DOMAIN;
                curFromItem.domainAttribute=new String(curObjectName);
        //                myname += "." + new String(curObjectName);
            }
        }
        curFromItem.name = myname;
    }

    /**
     * oracleObjectName -> OracleObjectName()
     */
    public void visit(ViewReference n) {
        n.oracleObjectName.accept(this);
        // this should occur only once
        if (bld.viewName!=null) {
            Util.exit("Trying to assign view name twice...weird!");
            
        }
        bld.viewName = new String(curObjectName);
    }

    /**
     * parameter -> Parameter()
     * nodeListOptional -> ( "," Parameter() )*
     */
    public void visit(ParameterList n) {
        n.parameter.accept(this);
        n.nodeListOptional.accept(this);
    }

    /**
     * nodeChoice -> <S_IDENTIFIER>
     *       | ( [ "+" | "-" ] <S_NUMBER> )
     */
    public void visit(NumOrID n) {
        n.nodeChoice.accept(this);
    }

    /**
     * nodeToken -> <S_IDENTIFIER>
     * nodeOptional -> [ [ "IN" ] [ "OUT" ] TypeDeclaration() [ ( ":=" | "DEFAULT" ) PlSqlExpression() ] ]
     */
    public void visit(Parameter n) {
        n.nodeToken.accept(this);
        n.nodeOptional.accept(this);
    }

    /**
     * plSqlExpressionList -> PlSqlExpressionList()
     */
    public void visit(Arguments n) {
        n.plSqlExpressionList.accept(this);
    }

    /**
     * selectWithoutOrder -> SelectWithoutOrder()
     * nodeOptional -> [ OrderByClause() ]
     * nodeOptional1 -> [ ForUpdateClause() ]
     */
    public void visit(SelectStatement n) {
        n.selectWithoutOrder.accept(this);
        n.nodeOptional.accept(this);
        n.nodeOptional1.accept(this);
    }

    /**
     * nodeToken -> "SELECT"
     * nodeOptional -> [ "ALL" | "DISTINCT" ]
     * selectList -> SelectList()
     * nodeOptional1 -> [ IntoClause() ]
     * fromClause -> FromClause()
     * nodeOptional2 -> [ WhereClause() ]
     * nodeOptional3 -> [ ConnectClause() ]
     * nodeOptional4 -> [ GroupByClause() ]
     * nodeOptional5 -> [ SetClause() ]
     */
    public void visit(SelectWithoutOrder n) {
        n.nodeToken.accept(this);
        n.nodeOptional.accept(this);
        n.selectList.accept(this);
        n.nodeOptional1.accept(this);
        n.fromClause.accept(this);
        n.nodeOptional2.accept(this);
        n.nodeOptional3.accept(this);
        n.nodeOptional4.accept(this);
        n.nodeOptional5.accept(this);
    }

    /**
     * nodeChoice -> "*"
     *       | SelectItem() ( "," SelectItem() )*
     */
    public void visit(SelectList n) {
        n.nodeChoice.accept(this);
        if (n.nodeChoice.which==0) {
            Util.exit("select * not handled yet");
            
        }
    }

    /**
     * nodeChoice -> OracleObjectName() ".*"
     *       | OracleObjectName() "." OracleObjectName() ".*"
     *       | SQLSimpleExpression() [ ColumnAlias() ]
     */
    public void visit(SelectItem n) {
        // initialize to test for NULLity later
        curObjectName=null;
        curConstant=null;
        curTableReference=null;
        curAttributeReference=null;
        curColumnAlias=null;

        n.nodeChoice.accept(this);
        if (n.nodeChoice.which<2) { // one of the first two choices
            curSelectItem = new tbSelectItem(curObjectName);
            curSelectItem.star=true;
        }
        else {
            // we have an sql-simple expression here. Can't be
            // arithmetic right now -- we'll go down all the way to
            // SQLPrimaryExpression
            if (curConstant!=null) {
                Util.exit("Constants in SELECT-clause not handled (yet)");
                
            }
            else {
                if (curTableReference==null) {
                    Util.exit("invalid select-item");
                    
                }
                else 
                    if (curAttributeReference!=null) 
                        curSelectItem = new 
                            tbSelectItem(curTableReference,
                                         curAttributeReference);
                    else 
                        curSelectItem = new tbSelectItem(curTableReference);
            }
        }
        if (curSelectItem!=null) {
            if (curColumnAlias!=null)
                curSelectItem.columnAlias=new String(curColumnAlias);
            bld.selectClause.addElement(curSelectItem);
        }
        else {
            Util.exit("no current select-item.. something's weird");
            
        }
    }

    /**
     * nodeToken -> <S_IDENTIFIER>
     */
    public void visit(ColumnAlias n) {
        curColumnAlias = n.nodeToken.toString();
        n.nodeToken.accept(this);
    }

    /**
     * nodeToken -> "("
     * createViewItem -> CreateViewItem()
     * nodeListOptional -> ( "," CreateViewItem() )*
     * nodeToken1 -> ")"
     */
    public void visit(CreateViewList n) {
        n.nodeToken.accept(this);
        n.createViewItem.accept(this);
        n.nodeListOptional.accept(this);
        n.nodeToken1.accept(this);
    }

    /**
     * oracleObjectName -> OracleObjectName()
     */
    public void visit(CreateViewItem n) {
        n.oracleObjectName.accept(this);
        bld.viewClause.addElement(new String(curObjectName));
    }

    /**
     * nodeToken -> "INTO"
     * intoItem -> IntoItem()
     * nodeListOptional -> ( "," IntoItem() )*
     */
    public void visit(IntoClause n) {
        n.nodeToken.accept(this);
        n.intoItem.accept(this);
        n.nodeListOptional.accept(this);
    }

    /**
     * nodeChoice -> ( <S_IDENTIFIER> [ "." <S_IDENTIFIER> ] )
     *       | ( <S_BIND> )
     */
    public void visit(IntoItem n) {
        n.nodeChoice.accept(this);
    }

    /**
     * nodeToken -> "FROM"
     * schemaSQLFromItem -> SchemaSQLFromItem()
     * nodeListOptional -> ( "," SchemaSQLFromItem() )*
     */
    public void visit(FromClause n) {
        n.nodeToken.accept(this);
        n.schemaSQLFromItem.accept(this);
        n.nodeListOptional.accept(this);
    }

    /**
     * nodeChoice -> SchemaSQL_RelationNames()
     *       | SchemaSQL_AttributeNames()
     *       | FromItem()
     */
    public void visit(SchemaSQLFromItem n) {
        // increase fromItemIndex by 1, that is the current fromItem
        fromItemIndex++;
        curFromItem = new tbFromItem();
        bld.fromClause.insertElementAt(curFromItem,fromItemIndex);
        n.nodeChoice.accept(this);
    }

    /**
     * tableReference -> TableReference()
     * nodeOptional -> [ TableAlias() ]
     */
    public void visit(FromItem n) {
        curFromItem.type = Builder.TUPLE;
        n.tableReference.accept(this);
        n.nodeOptional.accept(this);
    }


    /**
     * nodeToken -> <O_ARROW>
     * tableAlias -> TableAlias()
     */
    public void visit(SchemaSQL_RelationNames n) {
        curFromItem.type = Builder.RELNAME;
        n.nodeToken.accept(this);
        n.tableAlias.accept(this);
    }

    /**
     * tableReference -> TableReference()
     * nodeToken -> <O_ARROW>
     * columnAlias -> ColumnAlias()
     */
    public void visit(SchemaSQL_AttributeNames n) {
        curColumnAlias=null;
        curFromItem.type = Builder.ATTNAME;
        n.tableReference.accept(this);
        n.nodeToken.accept(this);
        n.columnAlias.accept(this);
        if (curColumnAlias!=null)
            curFromItem.alias=new String(curColumnAlias);
    }

    /**
     * nodeToken -> <S_IDENTIFIER>
     */
    public void visit(TableAlias n) {
        curFromItem.alias = n.nodeToken.toString();
        n.nodeToken.accept(this);
    }

    /**
     * nodeToken -> "WHERE"
     * sQLExpression -> SQLExpression()
     */
    public void visit(WhereClause n) {
        n.nodeToken.accept(this);
        n.sQLExpression.accept(this);
    }

    /**
     * nodeSequence -> ( [ "START" "WITH" SQLExpression() ] "CONNECT" "BY" SQLExpression() [ "START" "WITH" SQLExpression() ] )
     */
    public void visit(ConnectClause n) {
        n.nodeSequence.accept(this);
    }

    /**
     * nodeToken -> "GROUP"
     * nodeToken1 -> "BY"
     * sQLExpressionList -> SQLExpressionList()
     * nodeOptional -> [ "HAVING" SQLExpression() ]
     */
    public void visit(GroupByClause n) {
        n.nodeToken.accept(this);
        n.nodeToken1.accept(this);
        n.sQLExpressionList.accept(this);
        n.nodeOptional.accept(this);
    }

    /**
     * nodeChoice -> ( ( "UNION" [ "ALL" ] ) | "INTERSECT" | "MINUS" )
     * nodeChoice1 -> ( ( "(" SelectWithoutOrder() ")" ) | SelectWithoutOrder() )
     */
    public void visit(SetClause n) {
        n.nodeChoice.accept(this);
        n.nodeChoice1.accept(this);
    }

    /**
     * nodeToken -> "ORDER"
     * nodeToken1 -> "BY"
     * sQLSimpleExpression -> SQLSimpleExpression()
     * nodeOptional -> [ "ASC" | "DESC" ]
     * nodeListOptional -> ( "," SQLSimpleExpression() [ "ASC" | "DESC" ] )*
     */
    public void visit(OrderByClause n) {
        n.nodeToken.accept(this);
        n.nodeToken1.accept(this);
        n.sQLSimpleExpression.accept(this);
        n.nodeOptional.accept(this);
        n.nodeListOptional.accept(this);
    }

    /**
     * nodeToken -> "FOR"
     * nodeToken1 -> "UPDATE"
     * nodeToken2 -> "OF"
     * tableColumn -> TableColumn()
     * nodeListOptional -> ( "," TableColumn() )*
     */
    public void visit(ForUpdateClause n) {
        n.nodeToken.accept(this);
        n.nodeToken1.accept(this);
        n.nodeToken2.accept(this);
        n.tableColumn.accept(this);
        n.nodeListOptional.accept(this);
    }

    /**
     * sQLAndExpression -> SQLAndExpression()
     * nodeListOptional -> ( "OR" SQLAndExpression() )*
     */
    public void visit(SQLExpression n) {
        if (n.nodeListOptional.present()) {
            //We have an OR... not handled
            Util.exit("treeBuilder: SQL-OR not handled (yet)");
            
        }
        n.sQLAndExpression.accept(this);
        n.nodeListOptional.accept(this);
    }

    /**
     * sQLUnaryLogicalExpression -> SQLUnaryLogicalExpression()
     * nodeListOptional -> ( "AND" SQLUnaryLogicalExpression() )*
     */
    public void visit(SQLAndExpression n) {
        n.sQLUnaryLogicalExpression.accept(this);
        n.nodeListOptional.accept(this);
    }

    /**
     * nodeChoice -> ExistsClause()
     *       | ( [ "NOT" ] SQLRelationalExpression() )
     */
    public void visit(SQLUnaryLogicalExpression n) {
        if (n.nodeChoice.which == 0) {// ExistsClause
            Util.exit("treeBuilder: SQL-EXISTS not handled (yet)!");
            
        }
        n.nodeChoice.accept(this);
    }

    /**
     * nodeOptional -> [ "NOT" ]
     * nodeToken -> "EXISTS"
     * nodeToken1 -> "("
     * subQuery -> SubQuery()
     * nodeToken2 -> ")"
     */
    public void visit(ExistsClause n) {
        n.nodeOptional.accept(this);
        n.nodeToken.accept(this);
        n.nodeToken1.accept(this);
        n.subQuery.accept(this);
        n.nodeToken2.accept(this);
    }

    /**
     * nodeChoice -> ( "(" SQLExpressionList() ")" | ( [ "PRIOR" ] SQLSimpleExpression() ) )
     * nodeOptional -> ( SQLRelationalOperatorExpression() | ( SQLInClause() ) | ( SQLBetweenClause() ) | ( SQLLikeClause() ) | IsNullClause() )?
     */
    public void visit(SQLRelationalExpression n) {
        // when we evaluate these expression, all fromItems have been
        // created already.
        curConstraint =  new tbConstraint();
        curRelop=tbConstraint.ERROR;
        n.nodeChoice.accept(this);

        // now we have the left side of the expression. Only one of
        // the strings curTableReference or curConstant can be !=NULL
        if (curTableReference!=null) {
            tbFromItem tmp = bld.findFromItem(curTableReference);
            if (tmp==null) {
                Util.exit("tbFromItem not found for :"
                                   +curTableReference+". Bye bye.");
                
            }
            else {
                if (curAttributeReference!=null) 
                    curConstraint.left = 
                        new tbAttributeItem(tmp, curAttributeReference);
                else
                    curConstraint.left = tmp;
                curConstraint.type=tbConstraint.JOIN;
            }
        }
        else if (curConstant!=null) {
            curConstraint.left=new tbConst(curConstant);
            curConstraint.type=tbConstraint.LOCAL;
        }
        else {
            Util.exit("SQLRelationalExpression: both curTableReference and curConstant are set!");
            
        }
        curTableReference=curConstant=null;

        n.nodeOptional.accept(this);
        // this gives the right side of the expression, so we can do
        // the whole thing again for the right side now.
        
        if (curTableReference!=null) {
            tbFromItem tmp = bld.findFromItem(curTableReference);
            if (tmp==null) {
                Util.exit("tbFromItem not found for :"+curTableReference+". Bye bye.");
                
            }
            else
                if (curAttributeReference!=null) 
                    curConstraint.right=new tbAttributeItem(tmp,curAttributeReference);
                else
                    curConstraint.right = tmp;
            // if the left side doesn't already have a constant
            if (curConstraint.type!=tbConstraint.LOCAL) curConstraint.type=tbConstraint.JOIN;

        }
        else if (curConstant!=null) {
            curConstraint.right=new tbConst(curConstant);
            curConstraint.type=tbConstraint.LOCAL;
        }
        else {
            Util.exit("SQLRelationalExpression: both curTableReference and curConstant are set!");
            
        }
        curTableReference=curConstant=null;

        // now, add the relop to the constraint
        if (curRelop==tbConstraint.ERROR) {
            Util.exit("Invalid relational operator");
            
        }
        else
            curConstraint.operator = curRelop;

        // after the subtree is evaluated, add complete constraint to
        // appropriate from-items 
        boolean constraintAdded=false;
        tbItem tmp  = curConstraint.left;
        if (tmp instanceof tbFromItem) {
            ((tbFromItem)tmp).addConstraint(curConstraint);
            constraintAdded=true;
        }
        else if (tmp instanceof tbAttributeItem) {
            ((tbAttributeItem)tmp).fromItem.addConstraint(curConstraint);
            constraintAdded=true;
        }
        tmp  = curConstraint.right;
        if (tmp instanceof tbFromItem){
            ((tbFromItem)tmp).addConstraint(curConstraint);
            constraintAdded=true;
        }
        else if (tmp instanceof tbAttributeItem) {
            ((tbAttributeItem)tmp).fromItem.addConstraint(curConstraint);
            constraintAdded=true;
        }
        if(!constraintAdded) {
            bld.leftoverConstraints.addElement(curConstraint);
            System.out.println("SchemaSQLVisitor-Warning: standard SQL clause added. This may not work properly.");
        }
    }

    /**
     * sQLSimpleExpression -> SQLSimpleExpression()
     * nodeListOptional -> ( "," SQLSimpleExpression() )*
     */
    public void visit(SQLExpressionList n) {
        n.sQLSimpleExpression.accept(this);
        n.nodeListOptional.accept(this);
    }

    /**
     * relop -> Relop()
     * nodeChoice -> ( ( [ "ALL" | "ANY" ] "(" SubQuery() ")" ) | [ "PRIOR" ] SQLSimpleExpression() )
     */
    public void visit(SQLRelationalOperatorExpression n) {
        n.relop.accept(this);
        n.nodeChoice.accept(this);
    }

    /**
     * nodeOptional -> [ "NOT" ]
     * nodeToken -> "IN"
     * nodeToken1 -> "("
     * nodeChoice -> ( SQLExpressionList() | SubQuery() )
     * nodeToken2 -> ")"
     */
    public void visit(SQLInClause n) {
        n.nodeOptional.accept(this);
        n.nodeToken.accept(this);
        n.nodeToken1.accept(this);
        n.nodeChoice.accept(this);
        n.nodeToken2.accept(this);
    }

    /**
     * nodeOptional -> [ "NOT" ]
     * nodeToken -> "BETWEEN"
     * sQLSimpleExpression -> SQLSimpleExpression()
     * nodeToken1 -> "AND"
     * sQLSimpleExpression1 -> SQLSimpleExpression()
     */
    public void visit(SQLBetweenClause n) {
        n.nodeOptional.accept(this);
        n.nodeToken.accept(this);
        n.sQLSimpleExpression.accept(this);
        n.nodeToken1.accept(this);
        n.sQLSimpleExpression1.accept(this);
    }

    /**
     * nodeOptional -> [ "NOT" ]
     * nodeToken -> "LIKE"
     * sQLSimpleExpression -> SQLSimpleExpression()
     */
    public void visit(SQLLikeClause n) {
        n.nodeOptional.accept(this);
        n.nodeToken.accept(this);
        if (n.nodeOptional.present()) 
            curRelop=tbConstraint.NOTLIKE;
        else
            curRelop=tbConstraint.LIKE;
        n.sQLSimpleExpression.accept(this);
    }

    /**
     * sQLMultiplicativeExpression -> SQLMultiplicativeExpression()
     * nodeListOptional -> ( ( "+" | "-" | "||" ) SQLMultiplicativeExpression() )*
     */
    public void visit(SQLSimpleExpression n) {
        n.sQLMultiplicativeExpression.accept(this);
        if (n.nodeListOptional.present()) {
            Util.exit("Arithmetic expressions not handled (yet)");
            
        }
        n.nodeListOptional.accept(this);
    }

    /**
     * sQLExpotentExpression -> SQLExpotentExpression()
     * nodeListOptional -> ( ( "*" | "/" ) SQLExpotentExpression() )*
     */
    public void visit(SQLMultiplicativeExpression n) {
        n.sQLExpotentExpression.accept(this);
        if (n.nodeListOptional.present()) {
            Util.exit("Arithmetic expressions not handled (yet)");
            
        }
        n.nodeListOptional.accept(this);
    }

    /**
     * sQLUnaryExpression -> SQLUnaryExpression()
     * nodeListOptional -> ( "**" SQLUnaryExpression() )*
     */
    public void visit(SQLExpotentExpression n) {
        n.sQLUnaryExpression.accept(this);
        if (n.nodeListOptional.present()) {
            Util.exit("Arithmetic expressions not handled (yet)");
            
        }
        n.nodeListOptional.accept(this);
    }

    /**
     * nodeOptional -> [ "+" | "-" ]
     * sQLPrimaryExpression -> SQLPrimaryExpression()
     */
    public void visit(SQLUnaryExpression n) {
        if (n.nodeOptional.present()) {
            Util.exit("Arithmetic expressions not handled (yet)");
            
        }
        n.nodeOptional.accept(this);
        n.sQLPrimaryExpression.accept(this);
    }

    /**
     * nodeChoice -> "NULL"
     *       | FunctionCall()
     *       | OuterJoinExpression()
     *       | TableColumn()
     *       | <S_NUMBER>
     *       | <S_CHAR_LITERAL>
     *       | <S_BIND>
     *       | "(" SQLExpression() ")"
     */
    public void visit(SQLPrimaryExpression n) {
        // reset the tags
        curTableReference = null;
        curConstant = null;
        n.nodeChoice.accept(this);
        switch (n.nodeChoice.which) {
        case 3: // TableColumn()
            break;
        case 4: // S_NUMBER
            curConstant = new String(n.nodeChoice.choice.toString());
            break;
        case 5: // S_CHAR_LITERAL
            curConstant = new String(n.nodeChoice.choice.toString());
            // remove quotes
            curConstant = curConstant.substring(1,curConstant.length()-1);
            break;
        default:
            Util.exit("This primary expression is not handled: "+n.nodeChoice.choice.toString());
            
            break;
        }
        
    }

    /**
     * oracleObjectName -> OracleObjectName()
     * nodeOptional -> [ "." OracleObjectName() [ "." OracleObjectName() ] ]
     * nodeToken -> "("
     * nodeOptional1 -> [ [ "DISTINCT" | "ALL" ] ( SQLArguments() | "*" ) ]
     * nodeToken1 -> ")"
     */
    public void visit(FunctionCall n) {
        n.oracleObjectName.accept(this);
        n.nodeOptional.accept(this);
        n.nodeToken.accept(this);
        n.nodeOptional1.accept(this);
        n.nodeToken1.accept(this);
    }

    /**
     * sQLExpressionList -> SQLExpressionList()
     */
    public void visit(SQLArguments n) {
        n.sQLExpressionList.accept(this);
    }

    /**
     * oracleObjectName -> OracleObjectName()
     * nodeOptional -> [ "." OracleObjectName() [ "." OracleObjectName() ] ]
     * nodeToken -> "("
     * nodeToken1 -> "+"
     * nodeToken2 -> ")"
     */
    public void visit(OuterJoinExpression n) {
        n.oracleObjectName.accept(this);
        n.nodeOptional.accept(this);
        n.nodeToken.accept(this);
        n.nodeToken1.accept(this);
        n.nodeToken2.accept(this);
    }

    /**
     * selectWithoutOrder -> SelectWithoutOrder()
     */
    public void visit(SubQuery n) {
        n.selectWithoutOrder.accept(this);
    }

}
