package treeBuilder;

import schemaSQLParser.*;
import algebraTree.*;
import java.util.*;
import DSRG.MW.*;

/** the algebra tree builder. Parses the SchemaSQL query using package
 * schemaSQLParser and then builds an algebra tree using package
 * algebraTree.  */

public class Builder {

    private final static boolean DEBUG=false;

    public final static int TUPLE = 1;
    public final static int ATTNAME = 2;
    public final static int RELNAME = 3;
    public final static int DOMAIN = 4;
    public final static int ERROR=0;
    
    private static final int PIVOTNAME=1;
    private static final int  DATANAME=2;
    private static final int RELPIVOTNAME=3;
    
    /** whatever comes right after the view clause*/
    public String viewName=null;
    /** any items in the view clause... Strings, for the time being*/
    public Vector viewClause=new Vector(5);

    /** any item in the select clause... these are "tbSelectItem"s */
    public Vector selectClause=new Vector(5);

    public Vector fromClause=new Vector(5);

    // where-clause contains only the standard-SQL-where-statements
    // that did not get added to any from-variable. BUG 092200:
    // Apparently, this got forgotten as I'm now having this
    // problem. Will try again.
    public Vector leftoverConstraints=new Vector(5);

    final static String[] typeNames = new String[5];

    /** handles for the system's source and local DB connections, to
     * be used in tree-building
     */
    private DBManager sourceDB;
    private DBManager localDB;

    public Builder(DBManager sourceP, DBManager localP) {

        sourceDB = sourceP;
        localDB = localP;

        /* debugging purposes only! */
        typeNames[0]="ERROR";
        typeNames[1]="TUPLE";
        typeNames[2]="ATTNAME";
        typeNames[3]="RELNAME";
        typeNames[4]="DOMAIN";
    }


    /** kicks off creation of algebra tree. Must be called after the
     *   visitor is done 
     *   @return rootnode of the algebraTree
     */

    public QENode createAlgebraTree() {
        Vector flatList = makeFlattenList(fromClause);
        QENode currentNode,childNode;
        SchemaSQLOperator currentOp,childOp;
        Vector uniteRelNames = new Vector(2);
        String query;
        Vector constraints;
        tbConstraint con;
        boolean firstTime;
        Table tempTable;
        
        /** contains one QENode for each element in flatList */
        Vector joinList = new Vector(5); 

        /** associates variable names (from the from-clause) with the
         * corresponding SchemaSQLOperator-objects. This is necessary to later
         * find the automatically generated schema names at each node */
        Hashtable variableNames = new Hashtable(10);

        // there can be only one source IS right now. Take that
        // handle.
        

        tbFromItem curFromItem;
        for (int curFromItemIndex = 0 ; curFromItemIndex < flatList.size(); curFromItemIndex++) {
            // flatten each relation, i.e., build tree with unite,
            // fold, or SQL nodes...
            
            curFromItem = (tbFromItem)flatList.elementAt(curFromItemIndex);
            constraints = curFromItem.constraints;
            if (curFromItem.type == RELNAME) {
                // create UNITE-node
                // 1. query FST to retrieve relnames
                // 1a. build query. Go through constraints, take each
                // local constraint and add to select clause
                query = "select distinct table_name from fst";
                firstTime=true;
                for (int j=0; j<constraints.size(); j++) {
                    con = (tbConstraint)constraints.elementAt(j);
                    if (con.type==tbConstraint.LOCAL) {
                        if (firstTime) {
                            query += " where ";
                            firstTime = false;
                        }
                        else query+=" and ";
                        if (con.left instanceof tbConst)
                            query+=
                                "'"+((tbConst)con.left).getVALUE()+"'"+
                                con.getOperator()+
                                " table_name";
                        else
                            query+=
                                " table_name "+
                                con.getOperator()
                                +"'"+((tbConst)con.right).getVALUE()+"'";
                    }
                }
                // 1b. now send the query to sourceDB
                tempTable =sourceDB.executeViewQuery(query);
                uniteRelNames = tempTable.getColumn(0);
                if (DEBUG) 
                    System.out.println("uniteRelNames: "+uniteRelNames.toString());
                // 2. now create UNITE-node
                currentOp = new Unite(localDB, // UNITE is always a leaf
                                      sourceDB,// and writes into localDB
                                      uniteRelNames,//the relnames
                                      Util.getTempAttributeName()
                                      // and a new attribute name
                                      );
                currentNode = new QENode(currentOp);
            } // end of "==RELNAME"
            else if(curFromItem.type == TUPLE){
                query = "select * from "+
                    curFromItem.name;
                firstTime=true;
                for (int j=0; j<constraints.size(); j++) {
                    con = (tbConstraint)constraints.elementAt(j);
                    if (con.type==tbConstraint.LOCAL) {
                        if (firstTime) {
                            query += " where ";
                            firstTime = false;
                        }
                        else query+=" and ";
                        if (con.left instanceof tbConst)
                            query+=
                                "'"+((tbConst)con.left).getValue()+"'"+
                                con.getOperator()+
                                ((tbAttributeItem)con.right).attribute;
                        else
                            query+=
                                ((tbAttributeItem)con.left).attribute+
                                con.getOperator()+
                                "'"+((tbConst)con.right).getValue()+"'";
                    }
                }
                if (DEBUG) 
                    System.out.println("StandardSQLQuery: "+query);
                currentOp = new StandardSQL(localDB,sourceDB,query);
                currentNode = new QENode(currentOp);
            } // end of "==TUPLE"
            else if(curFromItem.type==ATTNAME) {
                System.err.println("WARNING: ATTNAMEs shouldn't be leaf nodes since we don't know how to handle this");
                Vector foldAttNames;
                constraints = curFromItem.constraints;
                // we have a Fold-leaf node.
                // curFromItem is an ATTNAME item
                query = "select column_name from fst where table_name='";
                query += curFromItem.name;
                query +="'";
                for (int j=0; j<constraints.size(); j++) {
                    con = (tbConstraint)constraints.elementAt(j);
                    if (con.type==tbConstraint.LOCAL) {
                        query+=" and ";
                        if (con.left instanceof tbConst)
                            query+=
                                "'"+((tbConst)con.left).getVALUE()+"'"+
                                con.getOperator()+
                                " column_name";
                        else
                            query+=
                                " column_name "+
                                con.getOperator()
                                +"'"+((tbConst)con.right).getVALUE()+"'";
                    }
                }
                tempTable = sourceDB.executeViewQuery(query);
                foldAttNames=tempTable.getColumn(0);
                
                // now we have all the foldAttNames and can create the
                // fold-node.
                
                if (DEBUG) System.out.println("foldAttNames(leaf): "
                                              +foldAttNames.toString());
                
                currentOp = new Fold(localDB, // always a leaf node
                                     sourceDB,
                                     foldAttNames, // see above
                                     curFromItem.name, // the input relname 
                                     Util.getTempAttributeName(), // pivot att
                                     Util.getTempAttributeName() // data att
                                     );
                
                currentNode = new QENode(currentOp);
            }
            else {
                currentNode=null; currentOp=null;// to make the compiler happy
                Util.exit("Invalid Variable in Flatten-List (Builder::createAlgebraTree)");
                
            }
            
            
            // 3. now, we have created one currentNode and need to
            // figure out what to do with it. If there is a
            // corresponding ATTNAME-variable, we need to Fold the
            // stuff, otherwise we are done and can add the current
            // node to the node list to be joined later.
            
            // we check if there is an ATTNAME item which matches our
            // current item in name or alias and if that's the case,
            // we add a fold-node on top. This will always be a
            // non-leaf node.
            
            // search from-clause for ATTNAME var, only one is allowed
            tbFromItem attVarItem=null;
            tbFromItem curItem;
            for (int i = 0 ; i<fromClause.size(); i++) {
                curItem = (tbFromItem)fromClause.elementAt(i);
                if (curItem.type!=ATTNAME) continue;
                if (curItem.name.equals(curFromItem.alias)) {
                    if (currentOp instanceof Unite) {
                        if (attVarItem==null)
                            attVarItem = (tbFromItem)fromClause.elementAt(i);
                        else {
                            Util.exit("two attVarItems with the same name-field, not handled. Builder.java");
                            
                        }
                    }
                    else // a fold node on top of a standard sql node
                        // implies that the attribute-name variable
                        // bases on a name, not alias. Otherwise,
                        // we're trying to find attribute names from a
                        // TUPLE-variable, which doesn't work (Xin
                        // found this one (072500)
                        Util.exit("Attribute-name variables cannot be based on Tuple variables");
                } // end of: curItem.name.equals(curFromItem.alias)
                else if (curItem.name.equals(curFromItem.name)) {
                    if (currentOp instanceof StandardSQL) {
                        if (attVarItem==null)
                            attVarItem = (tbFromItem)fromClause.elementAt(i);
                        else {
                            Util.exit("two attVarItems with the same name-field, not handled. Builder.java");
                            
                        }
                        // else impossible
                    }
                } // end of: curItem.name.equals(curFromItem.name)
            }
                
                
             // now, if attVarItem is not null there is a Fold node and
            // attVarItem has the right handle to an ATTNAME-variable
            // now comes the fold-operator
            
            if (attVarItem!=null) {
                
                // the UNITE or TUPLE node from above becomes a child node
                childNode=currentNode;
                childOp=currentOp;
                
                // now, we need to figure out the projection list of this
                // Fold-Operator
                
                // we resolve the table name that's coming from underneath
                // by looking at the original relname (TUPLE) or the first
                // name in the projection list in the UNITE node.
                
                Vector foldAttNames;
                constraints = attVarItem.constraints;
                
                query = "select column_name from fst where table_name='";
                
                if (childOp instanceof Unite) {
                    String firstEntry = (String)uniteRelNames.elementAt(0);
                    query +=firstEntry;
                }
                else if (childOp instanceof StandardSQL) { 
                    // we have a StandardSQL node above us
                    query +=curFromItem.name;
                }
                else { // something is weird
                    query=null; // make compiler happy
                    Util.exit("trying to add Fold-node to another Fold-Node.... confused 8-|");
                    
                }
                
                // handle constraints attached to an ATTNAME
                // variable. Since constraints are handled here, we'll
                // remove them from the leftoverConstraints list (which
                // contains the leftover constraints)
                
                query +="'";
                for (int j=0; j<constraints.size(); j++) {
                    con = (tbConstraint)constraints.elementAt(j);
                    if (con.type==tbConstraint.LOCAL) {
                        
                        
                        // remove from leftover list.
                        leftoverConstraints.removeElement(con);
                        // now construct where-clause
                        query+=" and ";
                        if (con.left instanceof tbConst)
                            query+=
                                "'"+((tbConst)con.left).getVALUE()+"'"+
                                con.getOperator()+
                                " column_name";
                        else
                            query+=
                                " column_name "+
                                con.getOperator()
                                +"'"+((tbConst)con.right).getVALUE()+"'";
                    }
                }
                tempTable = sourceDB.executeViewQuery(query);
                foldAttNames=tempTable.getColumn(0);
                
                // now we have all the foldAttNames and can create the
                // fold-node.
                
                if (DEBUG) System.out.println("foldAttNames: "
                                              +foldAttNames.toString());
                
                currentOp = new Fold(localDB, // always an internal node
                                     foldAttNames, // see above
                                     childNode.getRelName(), // the output
                                     // relname of the unite node
                                     Util.getTempAttributeName(), // pivot att
                                     Util.getTempAttributeName() // data att
                                     );
                
                currentNode = new QENode(currentOp);
                currentNode.addChild(childNode);
                
                if (attVarItem.alias!=null) {
                    if (variableNames.contains(attVarItem.alias)) 
                        Util.exit("hashing existing value...");
                    else
                        variableNames.put(attVarItem.alias,currentOp);
                }
                else {
                    if (variableNames.contains(attVarItem.name))
                        Util.exit("hashing existing value...");
                    else
                        variableNames.put(attVarItem.name,currentOp);
                }
            } // end of "if attVarItem!=null"
            
            // associating current node with an alias name so we can
            // later find it again based on the name it's serving if
            // we added a fold-node later, we want to use just that
            // fold node, even for underlying variables (e.g., from
            // the UNITE-node) since we're looking for the FLATTENED
            // table, which is always at the end of the
            // FLATTEN-mini-hierarchy... clear?? :)
            if (curFromItem.alias!=null) {
                if (variableNames.contains(curFromItem.alias))
                    Util.exit("hashing existing value...");
                else
                    variableNames.put(curFromItem.alias,currentOp);
            }
            else {
                if (variableNames.contains(curFromItem.name))
                    Util.exit("hashing existing value...");
                else
                    variableNames.put(curFromItem.name,currentOp);
            } 

            // we're done and can add this flattened relation (handle)
            // to the join list.
            
            joinList.addElement(currentNode);
            // These are the base object names that each flattened item
            // is based on. The actual flat-names are the
            // automatically generated names of the items in joinList
            // at the same index position.
            
        } // end of "for each relation to be flattened"

        // now we have a bunch of temporary relations in the localDB
        // which now have to be joined using a standard SQL query. For
        // this, we can query the relNames out of all nodes in
        // joinList and simply cross-product them. Then, we'll have to
        // worry about the projection list and the join conditions...

        // so here comes step 6 of Lakshmanan's algorithm: the
        // replacements of dom_i-variables (building of sigma and
        // pi-lists of the join expression)
        // steps 4 and 5 will be added later

        // select-clause contains all the "domain" names.
        // now make sense out of them

        if (DEBUG) System.out.println("variableNames hashtable: "+variableNames.toString());
        
        tbSelectItem curSelItem;
        tbFromItem firstFromItem,secondFromItem;
        Vector selectStringVector = new Vector();
        String selectString;
        for (int i=0; i < selectClause.size(); i++) {
            curSelItem = ((tbSelectItem)selectClause.elementAt(i));
            // matches the first and second names with a fromitem
            // alias, or, if there is no alias, with the fromitem name
            // and returns a handle to that fromItem
            firstFromItem = findFromItem(curSelItem.firstName);
            secondFromItem = findFromItem(curSelItem.secondName);

            if (firstFromItem==null) // a column name with no relation name
                selectString=curSelItem.firstName;
            else {
                if (DEBUG) System.out.println("Select-clause. Parsing type: "+firstFromItem.type);

                switch (firstFromItem.type) {
                case TUPLE: {
                    String tableName=
                        findFlatRelName(curSelItem.firstName,variableNames);
                    if (tableName==null) tableName =
                                             findFlatRelName(firstFromItem.name,variableNames);
                    if (secondFromItem==null ||
                        secondFromItem.type!=ATTNAME) {
                        // normal column 
                    
                        selectString =
                            tableName
                            + "." 
                            + curSelItem.secondName;
                    }
                    else { // special case: trying to have dataName in
                        // select-list. Here we have fold-nodes who are
                        // all in the join-list, so the procedure is
                        // similar to above
                        selectString =
                            tableName
                            + "." 
                            +findFlatAttName(curSelItem.secondName,variableNames,DATANAME);
                    }
                }
                break;
                case RELNAME: {
                    selectString =
                        findFlatRelName(curSelItem.firstName,variableNames)
                        +"."
                        +findFlatAttName2(curSelItem.firstName,variableNames,joinList);
                }
                case ATTNAME: {
                    selectString =
                        findFlatRelName(curSelItem.firstName,variableNames)
                        +"."
                        +findFlatAttName(curSelItem.firstName,variableNames,PIVOTNAME);
                }
                break;
                case DOMAIN: {
                    selectString = 
                        findFlatRelName(curSelItem.firstName,variableNames)
                        +"."
                        + firstFromItem.domainAttribute;
                }
                break;
                default:
                    selectString = null;
                    Util.exit("Select-clause parsing failed. Select item of wrong type");
                }
            }
            
            if (curSelItem.columnAlias!=null) 
                selectString += " as \""+curSelItem.columnAlias+"\"";
            selectStringVector.addElement(selectString);
            //            if (i<selectClause.size()-1) selectString+=",";
        }
        
        // so... select clause is done, except for the elements that
        // Unfold and Split nodes might add... 
        //now comes the WHERE-clause...  we want to replace each
        //variable with the intermediate table name that it is based
        //on.

        tbConstraint curConstraint;
        Vector constraintList = new Vector(10);
        String whereString = new String(" where "); 
        for (int i = 0 ; i < fromClause.size(); i++) {
            curFromItem = (tbFromItem)fromClause.elementAt(i);
            for (int j = 0 ; j < curFromItem.constraints.size(); j++) {
                curConstraint = (tbConstraint)curFromItem.
                    constraints.elementAt(j);
                // assumes that the equals method in tbConstraint works.
                if (curConstraint.type==tbConstraint.JOIN) 
                    if (!constraintList.contains(curConstraint))
                        constraintList.addElement(curConstraint);
            }
        }
        
        // now we have a list of all unique join-constraints. We can
        // now substitute the names in there similar to above.
        tbItem curItem;
        for (int i=0; i< constraintList.size();i++) {
            curConstraint = (tbConstraint)constraintList.elementAt(i);
            // each side of the constraint can be a fromitem, an
            // attribute-item, or a constant. If it's a fromitem, we
            // replace as above. For a constant, nothing changes. For
            // an att.reference-item, the left side is a tuple item,
            // so we can treat this similarly to regular tuple items
            // (in fact, I don't see the difference)

            // bad hack :) A two-loop
            for (curItem=curConstraint.left;;curItem=curConstraint.right) {
                if (curItem instanceof tbConst) {
                    // this case is impossible. can't have a join item with a constant in there
                    Util.exit("there is a constant in a join-clause... weird");
                    // old code. change 081600: whereString += ((tbConst)curItem).getValue();
                }
                else  // curItem is something else
                    if (curItem instanceof tbFromItem) {
                        curFromItem = (tbFromItem)curItem;
                        switch (curFromItem.type) {
                        case TUPLE: {
                            Util.exit("TUPLE item in WHERE-clause... weird");
                        }
                        break;
                        case RELNAME: {
                            whereString +=
                                findFlatRelName(curFromItem.alias,variableNames)
                                +"."
                                +findFlatAttName2(curFromItem.alias,variableNames,joinList);
                        }
                        break;

                        case ATTNAME: {
                            whereString +=
                                findFlatRelName(curFromItem.name,variableNames)
                                +"."
                                
                                +findFlatAttName(curFromItem.name,variableNames,PIVOTNAME);
                        }
                        break;
                        case DOMAIN: {
                            whereString += 
                                findFlatRelName(curFromItem.name,variableNames)
                                +"."
                                + curFromItem.domainAttribute;
                        }
                        break;
                        default:
                            Util.exit("where-clause parsing failed. where-item of wrong type");
                        }
                    }
                    else if (curItem instanceof tbAttributeItem) {
                        // this is like the "TUPLE" case above
                        tbAttributeItem curAttItem=(tbAttributeItem)curItem;
                        String tableName=
                            findFlatRelName(curAttItem.fromItem.name,variableNames);
                        if (tableName==null) tableName =
                                                 findFlatRelName(curAttItem.fromItem.alias,variableNames);
                        secondFromItem = findFromItem(curAttItem.attribute);
                        if (secondFromItem==null ||
                            secondFromItem.type!=ATTNAME) {
                            // normal column 
                            
                            whereString +=
                                tableName
                                + "." 
                                + curAttItem.attribute;
                        }
                        else { // special case: trying to have dataName in
                            // select-list. Here we have fold-nodes who are
                            // all in the join-list, so the procedure is
                            // similar to above
                            whereString +=
                                tableName
                                + "." 
                                +findFlatAttName(curAttItem.attribute,variableNames,DATANAME);
                        }
                        
                    }
                    else Util.exit("curItem of wrong type. confused....");
                
                if (curItem==curConstraint.left)
                    // we now add the operator
                    whereString += curConstraint.getOperator();
                else break;
                // instead of for-loop break condition: if we're at the right side, we're done
            } // end of for-loop
            if (i < constraintList.size()-1) whereString += " and ";
        } // end of iteration through list of contraints
        
        
        // Unfold and Split nodes.... have to be done before the
        // select clause since they affect the projection list of the
        // middle (standard-sql) node.

        // Unfold first: we need to determine the two parameters of
        // the unfold operation. The pivotName is coming from the
        // entry in the createView clause, the data name from its
        // corresponding entry in the select clause. So let's first
        // parse the createView Clause.

        // we have to decide which of the elements in the view clause
        // may be a variable. Only one may occur, and that one could
        // could be from any alias name in the fromList.

        // only full name matches, i.e., no element in the create view
        // clause can be like A.B !!!
        tbFromItem unfoldItem = null, tmpUnfoldItem;
        String tmpName; int unfoldIndex=-1;
        String unfoldPivotName=null,unfoldDataName=null;
        for(int i = 0 ; i < viewClause.size(); i++) {
            tmpName = (String)viewClause.elementAt(i);
            if (tmpName.indexOf(".")>=0) 
                Util.exit("dots (.) in CREATE-VIEW clause not allowed");
            tmpUnfoldItem = findFromItem(tmpName);
            if (tmpUnfoldItem!=null) {// we found a candidate
                if (unfoldItem==null) {// it's the first candidate
                    unfoldItem = tmpUnfoldItem;
                    unfoldIndex = i;
                }
                else
                    Util.exit("More than one domain variable in CREATE-VIEW clause!");
            }
        }



        if (unfoldItem!=null) {// we need a Unfold-node
            // unfoldItem now contains the FromItem that corresponds
            // to the variable that we're using in the CREATE-VIEW-clause
            // unfoldIndex is its position, which should be the
            // position of the corresponding select-Item in the
            // selectClause

            switch  (unfoldItem.type) {
            case RELNAME:
                // looking for the pivotname of the unite node
                unfoldPivotName = 
                    findFlatRelName(unfoldItem.alias,variableNames)
                    +"."+
                    findFlatAttName2(unfoldItem.alias,variableNames,joinList);
                break;
            case ATTNAME:
                // the pivotname of the fold node
                unfoldPivotName = 
                    findFlatRelName(unfoldItem.alias,variableNames)
                    +"."+
                    findFlatAttName(unfoldItem.alias,variableNames,PIVOTNAME);
                break;
            case DOMAIN:
                // some fixed name from a StandardSQL node (?)
                String tmpFlatRelName = findFlatRelName(unfoldItem.name,variableNames);
                if (tmpFlatRelName==null) 
                    Util.exit("Please use relation variable for following from-Item: "+unfoldItem.toString());
                unfoldPivotName = 
                    tmpFlatRelName
                    +"."+
                    unfoldItem.domainAttribute;
                // the attribute name from the select-entry (should be
                // a constant attribute name)
                break;
            default:
                Util.exit("unfoldItem of wrong type:"+unfoldItem.type);
            }
            
            tmpName = ((tbSelectItem)selectClause.elementAt(unfoldIndex)).secondName;
            if (tmpName==null) // no table name given in the select clause
                Util.exit("please provide a relation name for dynamic select-clause "+((tbSelectItem)selectClause.elementAt(unfoldIndex)).toString()+" This case is not handled.");
            // try to find that name in the variableNames (then it's dynamic)
            unfoldDataName = findFlatAttName(tmpName,variableNames,DATANAME);
            // otherwise, just take the attribute name
            if (unfoldDataName==null)
                unfoldDataName=tmpName;

            // add the relation name
            tmpName = ((tbSelectItem)selectClause.elementAt(unfoldIndex)).firstName;
            // determine relation name for this select item
            String unfoldDataRelName = findFlatRelName(tmpName,variableNames);
            // if not found directly, try to dereference through the from-clause
            if (unfoldDataRelName==null) 
                unfoldDataRelName = findFlatRelName(findFromItem(tmpName).name,variableNames);
            unfoldDataName = unfoldDataRelName + "." + unfoldDataName;

        // now we can add those unfold  attributes to the selectString
            if (!selectStringVector.contains(unfoldPivotName))
                selectStringVector.addElement(unfoldPivotName);
            if (!selectStringVector.contains(unfoldDataName))
                selectStringVector.addElement(unfoldDataName);
        }


        // now we'll try to find information on a split node and
        // determine the pivotname for it
        tbFromItem splitItem = null;
        String splitPivotName=null;
        splitItem = findFromItem(viewName);

        // if we have a split node, determine pivotname
        if (splitItem != null) {
            
            switch  (splitItem.type) {
            case RELNAME:
                // looking for the pivotname of the unite node
                splitPivotName = 
                    findFlatRelName(splitItem.alias,variableNames)
                    +"."+
                    findFlatAttName2(splitItem.alias,variableNames,joinList);
                break;
            case ATTNAME:
                // the pivotname of the fold node
                splitPivotName = 
                    findFlatRelName(splitItem.alias,variableNames)
                    +"."+
                    findFlatAttName(splitItem.alias,variableNames,PIVOTNAME);
                break;
            case DOMAIN:
                // some fixed name from a StandardSQL node (?)
                String tmpFlatRelName = findFlatRelName(splitItem.name,variableNames);
                if (tmpFlatRelName==null) 
                    tmpFlatRelName = findFlatRelName(findFromItem(splitItem.name).name,variableNames);
                if (tmpFlatRelName==null)  // if it's still null
                    Util.exit("Please use relation variable for following from-Item: "+splitItem.toString());
                splitPivotName = 
                    tmpFlatRelName
                    +"."+
                    splitItem.domainAttribute;
                // the attribute name from the select-entry (should be
                // a constant attribute name)
                break;
            default:
                Util.exit("splitItem of wrong type:"+splitItem.type);
            }

        // now we can add the split  attribute to the selectString
            if (!selectStringVector.contains(splitPivotName))
                selectStringVector.addElement(splitPivotName);
        }

        // this code will be completely unmaintainable.... :(

        

        // and now we can continue with building the nodes...

        // create the select-string out of the vector:

        selectString = (String)selectStringVector.elementAt(0);;
        for (int i=1; i < selectStringVector.size();i++) 
            selectString += "," +(String)selectStringVector.elementAt(i);

        // now construct the query
        query="select "
            + selectString
            +" from ";
        for (int i=0; i<joinList.size(); i++) {
            query += ((QENode)joinList.elementAt(i)).getRelName();
            if (i<joinList.size()-1) query += ",";
        }


        //BUG  (found  092200): at  this  point,  we're missing  local
        //conditions that  do not correspond to a  flat table. Example
        //in airline.sql. Not sure how to  fix that - we would have to
        //identify such conditions. 
        
        if (leftoverConstraints.size()>0) {
            // now we have the leftoverConstraints with the orphan
            // constraints. Those strings in there should now be
            // translated into the proper temp-names. However, after a
            // month I forgot how that is done :)
            tbConstraint tmpConstr=null;
            String flatAttributeName=null;
            for (int i = 0; i < leftoverConstraints.size();i++) {
                // if the whereString has nothing in it, suppress
                // 'and', otherwise, put it there
                if (i>=1 || constraintList.size()>0) 
                    whereString += " and ";
                tmpConstr=(tbConstraint)leftoverConstraints.elementAt(i);
                if (tmpConstr.type!=tbConstraint.LOCAL)
                    Util.exit("somehow a join constraint got added to leftoverConstraints ... weird");
                if (tmpConstr.left instanceof tbConst) {
                    // not sure why I put DATANAME here, there is no
                    // reason that I understand why this should always
                    // be a DATANAME
                    flatAttributeName=findFlatAttName(((tbAttributeItem)tmpConstr.right).attribute,variableNames,DATANAME);
                    whereString+=
                        "'"+((tbConst)tmpConstr.left).getValue()+"'"+
                        tmpConstr.getOperator()+
                        flatAttributeName;
                }
                else {
                    flatAttributeName=findFlatAttName(((tbAttributeItem)tmpConstr.left).attribute,variableNames,DATANAME);
                    whereString+=
                        flatAttributeName+
                        tmpConstr.getOperator()+
                        "'"+((tbConst)tmpConstr.right).getValue()+"'";
                }
            }
        }

        query += (constraintList.size()>0 || leftoverConstraints.size()>0)?whereString:""; 

        if (DEBUG) System.out.println("Join Query: "+query);

        

        // create node and link to tree
        currentOp = new StandardSQL(localDB,query);
        
        currentNode = new QENode(currentOp);
        for (int i=0; i<joinList.size(); i++) {
            currentNode.addChild((QENode)joinList.elementAt(i));
        }
        
        // phew! :) 

        
        if (unfoldItem!=null) {
            // this is the second part of the Unfold-treatment. Comes
            // after the StandardSQL-node is created

            childNode = currentNode; // push current Node

            // the pivot names have to go into the node without table
            // names attached...
            if (unfoldPivotName.indexOf(".")>=0)
                unfoldPivotName = unfoldPivotName.substring(unfoldPivotName.indexOf(".")+1);
            if (unfoldDataName.indexOf(".")>=0)
                unfoldDataName = unfoldDataName.substring(unfoldDataName.indexOf(".")+1);
            currentOp = new Unfold(localDB, // UNFOLD is never a leaf
                                   childNode.getRelName(),
                                   unfoldPivotName,//ufPar1
                                   unfoldDataName//ufPar2
                                   );
            currentNode = new QENode(currentOp);
            currentNode.addChild(childNode);
        }
        
        // and finally - the split node (if we have one)
        if (splitItem!=null) {


            childNode = currentNode; // push current Node

            // the pivot names have to go into the node without table
            // names attached...
            if (splitPivotName.indexOf(".")>=0)
                splitPivotName = splitPivotName.substring(splitPivotName.indexOf(".")+1);
            currentOp = new Split(localDB, // SPLIT is never a leaf
                                   childNode.getRelName(),
                                   splitPivotName//splitPar
                                   );
            currentNode = new QENode(currentOp);
            currentNode.addChild(childNode);
        }
        else // there is no split item, so there could be a viewname
             // that is intended as the name of the output relation
            {
                if (viewName!=null)
                    currentNode.setRelName(viewName);
            }


        // currentNode should now be the rootnode of the complete algebra tree.
        return currentNode;
    }        


    /** this method finds all fromitems which should be flattened
     * (i.e., base relations.)  in a later version, the sources should
     * be read from a configuration file rather than
     * hardcoded. Currently, all sources must even exist in the same
     * source database (since we can't otherwise decide which source
     * comes from which database)
     * <BR> This will keep RELNAME, ATTNAME, and TUPLE variables/
     * @param the fromClause from this class, included as parameter
     * for clarity */

    private Vector makeFlattenList(Vector fromClauseP) {
        Vector result = (Vector)fromClauseP.clone(); // copies only
        // references!!!, not the actual constraints...
        String reduced,orig;
        tbFromItem tmpItem;

        // remove all elements which are based on non-base tables (i.e.,
        // whose names are other items' aliases)
        for (int i=0; i<result.size(); i++) // result.size() gets
                                            // recomputed every time
            for (int j=0; j<fromClauseP.size(); j++) {
                tmpItem = (tbFromItem)result.elementAt(i);
                reduced = tmpItem.name;
                orig= ((tbFromItem)fromClauseP.elementAt(j)).alias;
                if (reduced!=null && reduced.equals(orig)) {
                    if (tmpItem.constraints.size()!=0)
                        saveLocalConstraints(tmpItem);
                    result.removeElementAt(i--); // adjust i, too
                    break;  // remove only once
                }
            }

        // now remove all elements which are not RELNAME, ATTNAME, or
        // TUPLE
        //(i.e., DOMAIN vars :) )
        // NO!! (koeller 072500): ATTNAMEs must be removed. If they're
        // leaf nodes, we'll copy their base table to the local DB!!
        int curType;
        for (int i=0; i<result.size(); i++) {
            tmpItem = (tbFromItem)result.elementAt(i);
            curType = tmpItem.type;
            if ( (curType!=RELNAME) && (curType!=TUPLE)/*&& (curType!=ATTNAME)*/) {
                if (tmpItem.constraints.size()!=0)
                    saveLocalConstraints(tmpItem);
                result.removeElementAt(i--);
            }
        }    
        if (DEBUG) System.out.println("flatten-list: "+result);
        return result;
    }
    
    /** this stores any local constraint in the parameter from-Item in
        the global variable leftoverConstraints. */
    private void saveLocalConstraints(tbFromItem frmItem) {
        Vector tmpConstr = frmItem.constraints;
        tbConstraint oneConstr;

        for (Enumeration enum=tmpConstr.elements();enum.hasMoreElements();) {
            oneConstr = (tbConstraint)enum.nextElement();
            if (oneConstr.type==tbConstraint.LOCAL)
                if (!leftoverConstraints.contains(oneConstr))
                    leftoverConstraints.addElement(oneConstr);
        }
    }

    /** finds the fromItem with the given tableName as alias (if
        present) or name (if no alias) */
    public tbFromItem findFromItem(String tableName) {
        tbFromItem tmp;
        if (tableName==null) return null;
        for(Enumeration enum=fromClause.elements();
            enum.hasMoreElements();
            ) {
            tmp = (tbFromItem)enum.nextElement();
            if (tmp.alias==null) {
                if(tmp.name.equalsIgnoreCase(tableName))
                    return tmp;
            }
            else
                if (tmp.alias.equalsIgnoreCase(tableName))
                    return tmp;
        }
        return null;
    }


    /** searches a value in the variableNames hashtable and returns
     * the relname of the node that is hashed there */
    private String findFlatRelName(String search, Hashtable varNames)
    {
        SchemaSQLOperator tmp = (SchemaSQLOperator)varNames.get(search);
        if (tmp==null) {
            if (DEBUG) System.out.println("findFlatRelName: "+search+" not found in "+varNames.toString());
            return null;
        }
        return tmp.getRelName();
    }

    /** searches a value in the variableNames hashtable and returns
     * the pivot (type==TRUE) or data name (type==FALSE) of the node
     * there */
    private String findFlatAttName(String search, Hashtable varNames,
                                   int type) {
        SchemaSQLOperator x =
            (SchemaSQLOperator)varNames.get(search);
        // this is a bad hack. The problem is that a hashtable casts
        // an operator up to "Object", so I have to cast it down and
        // can't access the subclass methods anymore unless I cast to the
        // right class.... I should ask Xin about this (koeller 072500)
        if (x==null) return null;
        
        Fold tmpF=null; Unite tmpU=null;
        if (x instanceof Unite) {
            tmpU = (Unite)x;
            if (type==PIVOTNAME)
                return tmpU.getPivotName();
            else
                return null;
        }
        else 
            if (x instanceof Fold) {
                tmpF = (Fold)x;
                if (type==PIVOTNAME)
                    return tmpF.getPivotName();
                else
                    return tmpF.getDataName();
            }
            else return null;
        
    }

    /** another hack :( This function searches for an element in the
        join list and then goes to a Unite node that is its child. The
        purpose is that for relname items we need the pivot att name
        from the UNITE-ode, not the fold node above. I forgot about
        that.... Works only for cases in which the Unite-Operator is a
        direct child of the Fold-Operator... The case of two nested
        Fold-Operators is never handled in this implementation. Bug
        083100: this doesn't work if there is no fold-node at all. So
        since I forgot how this works - I'll just introduce a special
        case that omits the getChild() and hope that it miraculously
        works. God bless software-engineering...*/
    private String findFlatAttName2(String search, Hashtable varNames, Vector joinList) {
        Unite tmp=null;
        SchemaSQLOperator y;
        SchemaSQLOperator x =
            (SchemaSQLOperator)varNames.get(search);
        for (int i = 0 ; i < joinList.size() ; i++) {
            y=((QENode)joinList.elementAt(i)).getOperator();
            if (x==y) {
                // we got our candidate. 
                QENode tmpNode = (QENode)joinList.elementAt(i);
                if (tmpNode.getChild(0)==null)
                    tmp = (Unite)tmpNode.getOperator();
                else
                    tmp = (Unite)((QENode)joinList.elementAt(i)).getChild(0).getOperator();
            }
        }
        return tmp==null?null:tmp.getPivotName();
    }


    public String toString() {
        String result = new String();
        if (viewName!=null) {
            result += "ViewName: "+viewName +"\n";
            result += "View clause: "+viewClause.toString() +"\n";
        }
        result += "select: "+selectClause.toString()+"\n";
        result += "from: "+fromClause.toString();
     
        return result;
    }        
}
